"""Function to create map plots of the results with help of
the geopanda library (http://geopandas.org)
"""
import os
import logging
import copy
import math
import numpy as np
import geopandas as gpd
import pandas as pd
import palettable
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.patches import Circle
from matplotlib.colors import LinearSegmentedColormap

from energy_demand.basic import basic_functions
from energy_demand.technologies import tech_related
from energy_demand.read_write import write_data
#matplotlib.use('Agg') # Used to make it work in linux

def get_reasonable_bin_values_II(
        data_to_plot,
        nr_of_intervals
    ):
    """Get reasonable bins for 0....max value
    """
    def round_down(num, divisor):
        return num - (num%divisor)

    max_val = float(max(data_to_plot))
    min_val = float(min(data_to_plot))

    bin_width = max_val / nr_of_intervals

    bins = [bin_width * i for i in range(nr_of_intervals)]

    return bins

def get_reasonable_bin_values(
        data_to_plot,
        increments=10
    ):
    """Get reasonable bin values
    """
    def round_down(num, divisor):
        return num - (num%divisor)

    max_val = max(data_to_plot)
    min_val = min(data_to_plot)

    # Positive values
    if min_val >= 0:

        # Calculate number of classes
        classes = max_val / increments

        # Round down and add one class for larger values
        nr_classes = round_down(classes, increments) + 1

        #if nr_classes > 9:
        #    raise Exception("Nr of classes is too big")

        # Classes
        min_class = round_down(min_val, increments)                              # Minimum class
        max_class = round_down(max_val, increments) + increments + increments    # Maximum class

        if min_class == 0:
            min_class = increments

        # Bin with classes
        bins = list(range(int(min_class), int(max_class), int(increments)))

        # ---
        # Test that maximum 9 classes
        # ---
        if len(bins) > 8:
            print("Nr of bins: " + str(len(bins)))
            print(bins)
            raise Exception("Too many bin classes defined" + str(len(bins)))

    else:
        logging.info("Neg", flush=True)
        #lager negative values

        # must be of uneven length containing zero
        largest_min = abs(min_val)
        largest_max = abs(max_val)

        nr_min_class = round_down(abs(min_val), increments) / increments

        if max_val < 0:
            nr_pos_classes = 0
        else:
            nr_pos_classes = round_down(max_val, increments) / increments + 1

        # Number of classes
        #symetric_value = nr_of_classes_symetric * increments
        min_class_value = int(nr_min_class * -1) * increments
        max_class_value = int(nr_pos_classes) * increments

        # Negative classes
        if min_class_value / increments == 0:
            neg_classes = [increments * -1]
        elif min_class_value / increments == 1:
            neg_classes = [increments * -1]
        else:
            neg_classes = list(range(min_class_value, 0, increments))

        if max_class_value == 0:
            pos_classes = []
        elif max_class_value / increments == 0:
            pos_classes = [increments]
        elif max_class_value / increments == 1:
            pos_classes = [increments] #only one class
        else:
            pos_classes = list(range(increments, max_class_value + increments, increments))

        bins = neg_classes + [0] + pos_classes

        # ---
        # Test that maximum 9 classes
        # ---
        if len(bins) > 16:
            print("Nr of bins: " + str(len(bins)))
            print(bins)
            raise Exception("Too many bin classes defined" + str(len(bins)))

    return bins

def user_defined_classification(
        bins,
        min_value,
        max_value,
        lad_geopanda_shp,
        field_to_plot,
        legend_title,
        color_palette,
        color_prop,
        placeholder_zero_color,
        color_zero=False,
        color_list=False
    ):
    """Generate plot with own classification of choropleth maps

    Arguments
    ---------
    bins : list
        List with own borders for classification
    min_value : float
        Mininum data value
    lad_geopanda_shp : geopanda dataframe
        Figure pandaframe
    field_to_plot : str
        Name of figure to plot
    axes : axes
        Axis of figure
    color_palette : list
        List name of color scheme
    color_prop : str
        If wither sequential or qualitative colors

    Info
    ----
    Source: https://stackoverflow.com/questions/41783090/plotting-a-
    choropleth-map-with-geopandas-using-a-user-defined-classification-s
    """
    # --------------
    # Color paletts
    # --------------
    if color_prop == 'sequential':
        color_list = getattr(palettable.colorbrewer.sequential, color_palette).hex_colors
    elif color_prop == 'qualitative':
        color_list = getattr(palettable.colorbrewer.qualitative, color_palette).hex_colors
    elif color_prop == 'user_defined':
        color_list = color_list

    # Shorten color list
    if min(bins) > 0:
        color_list = color_list[:len(bins)+1]
    else:
        color_list = color_list

    # ------------------------------------------------------
    # Reclassify
    # ------------------------------------------------------
    reclass_lad_geopanda_shp, cmap = re_classification(
        lad_geopanda_shp,
        bins,
        color_list,
        field_to_plot,
        color_zero,
        placeholder_zero_color)

    # ------------------------------------------------------
    # Legend and legend labels
    # ------------------------------------------------------
    legend_handles = get_legend_handles(bins, color_list, color_zero, min_value, max_value)
    '''logging.info("bins fff")
    logging.info(bins)
    logging.info(color_list)

    legend_handles = []
    small_number = 0.01 # Small number for plotting corrrect charts

    if max(bins) >= 0:
        bins.append(999) # Append dummy last element for last class
    else:
        pass

    for bin_nr, bin_entry in enumerate(bins):
        if bin_nr == 0: #first bin entry

            if bin_entry < 0:
                label_patch = "> {} (min {})".format(bin_entry, min_value)

                if min_value > bin_entry:
                    print("Classification boundry is not clever for low values")
            else:
                label_patch = "< {} (min {})".format(bin_entry, min_value)
        elif bin_nr == len(bins)- 1: # -1 means that last bin entry
            ###label_patch = "> {} (max {})".format(bins[-2], max_value)
            label_patch = "> {} (max {})".format(bins[-2], max_value)

            if max_value < bin_entry:
                print("Classification boundry is not clever for low values")
        else:
            # ----------------------------
            # Add zero label if it exists
            # ----------------------------
            if bins[bin_nr - 1] == 0:
                patch = mpatches.Patch(
                    color=color_zero,
                    label=str("0"))
                legend_handles.append(patch)
            else:
                pass

            # ----------------------------
            # Other other labels
            # ----------------------------
            if bin_entry < 0:
                label_patch = "{}  ―  {}".format(bins[bin_nr - 1], bin_entry - small_number)
            else:
                if bins[bin_nr - 1] == 0:
                    label_patch = "{}  ―  {}".format(bins[bin_nr - 1] + small_number, bin_entry - small_number)
                else:
                    label_patch = "{}  ―  {}".format(bins[bin_nr - 1], bin_entry - small_number)

        logging.info("------label_patch: {}  {}  {}  {}".format(label_patch, bin_entry, bins, color_list))

        patch = mpatches.Patch(
            color=color_list[bin_nr],
            label=str(label_patch))

        legend_handles.append(patch)
    '''

    plt.legend(
        handles=legend_handles,
        title=legend_title,
        prop={'size': 5},
        loc='upper center',
        bbox_to_anchor=(0.5, -0.05),
        frameon=False)

    return reclass_lad_geopanda_shp, cmap

def get_legend_handles(bins, color_list, color_zero, min_value, max_value):
    """
    """
    legend_handles = []
    small_number = 0.0001 # Small number for plotting corrrect charts

    if max(bins) >= 0:
        bins.append(999) # Append dummy last element for last class
    else:
        pass

    for bin_nr, bin_entry in enumerate(bins):
        if bin_nr == 0: #first bin entry

            if bin_entry < 0:
                label_patch = "> {} (min {})".format(bin_entry, min_value)

                if min_value > bin_entry:
                    #print("Classification boundry is not clever for low values")
                    pass
            else:
                label_patch = "< {} (min {})".format(bin_entry, min_value)
        elif bin_nr == len(bins)- 1: # -1 means that last bin entry
            label_patch = "> {} (max {})".format(bins[-2], max_value)

            if max_value < bin_entry:
                #print("Classification boundry is not clever for low values")
                pass
        else:
            # ----------------------------
            # Add zero label if it exists
            # ----------------------------
            if bins[bin_nr - 1] == 0:
                patch = mpatches.Patch(
                    color=color_zero,
                    label=str("0"))

                legend_handles.append(patch)
            else:
                pass

            # ----------------------------
            # Other other labels
            # ----------------------------
            if bin_entry < 0:
                label_patch = "{}  ―  {}".format(bins[bin_nr - 1], bin_entry - small_number)
            else:
                if bins[bin_nr - 1] == 0:
                    label_patch = "{}  ―  {}".format(bins[bin_nr - 1] + small_number, bin_entry - small_number)
                else:
                    label_patch = "{}  ―  {}".format(bins[bin_nr - 1], bin_entry - small_number)

        logging.info("------label_patch: {}  {}  {}  {}".format(label_patch, bin_entry, bins, color_list))

        patch = mpatches.Patch(
            color=color_list[bin_nr],
            label=str(label_patch))

        legend_handles.append(patch)

    return legend_handles

def add_simple_legend(bins, color_list, color_zero, patch_form='rectangle'):
    """Add legend without giving the intervals
    """
    legend_handles = []

    bin_nr = 0
    for cnt, bin_value in enumerate(bins):

        if cnt == 0: # first entry
            if bin_value == 0:
                label = str(bin_value)
            else:
                label = str("< {}".format(bins[cnt + 1]))
        elif cnt == len(bins) - 1: #last entry
            label = str("> {}".format(bins[cnt - 1]))
        else:
            label = str(bin_value)

        if bin_value == 0:
            color = color_zero
        else:
            color = color_list[bin_nr]
            bin_nr += 1

        if patch_form == 'rectangle':
            patch = mpatches.Patch(
                color=color,
                label=label)
        elif patch_form == 'circle':
            patch = Circle(
                xy=(0,0),
                radius=1,
                color=color,
                label=label)
        else:
            raise Exception("wrong patch_form")

        legend_handles.append(patch)

    return legend_handles

def bin_mapping(
        value_to_classify,
        nr_of_classes,
        class_bins,
        placeholder_zero_color
    ):
    """Maps values to a bin.
    The mapped values must start at 0 and end at 1.

    Arguments
    ---------
    value_to_classify : float
        Value to classify
    class_bins : list
        Bins to use for classification

    Returns
    -------
    classified value
    """
    round_digits = 4

    value_to_classify = round(value_to_classify, round_digits)

    # Treat -0 and 0 the same
    if value_to_classify == 0 or value_to_classify == -0:
        #value_to_classify = 0

        # get position of zero value
        '''for idx, bound in enumerate(class_bins):
            if value_to_classify == bound:
                return idx / (len(class_bins) - 1.0)'''
        return placeholder_zero_color

    for idx, bound in enumerate(class_bins):
        if value_to_classify < bound:
            #return idx / (len(class_bins) - 1.0)
            #logging.info("TT " + str(idx / nr_of_classes - 1.0))
            return idx / (nr_of_classes - 1.0)

def re_classification(
        lad_geopanda_shp,
        bins,
        color_list,
        field_to_plot,
        color_zero,
        placeholder_zero_color
    ):
    """Reclassify according to user defined classification

    Arguments
    ----------
    lad_geopanda_shp : dataframe
        Shapefile geopanda dataframe
    bins : list
        Classification boders
    color_list : list
        List with colors for every category
    field_to_plot : str
        Name of figure to plot
    color_zero : str
        Color for zeros
    placeholder_zero_color : float
        Number to assign for zero values
    """
    # Create the list of bin labels and the
    # list of colors corresponding to each bin
    if max(bins) <=0:
        nr_of_classes = int(len(bins)) + 1 #only negative classes
    elif min(bins) >= 0:
        nr_of_classes = int(len(bins)) + 1 #only negative classes
    else:
        nr_of_classes = int(len(bins) -1) + 2 #class on top and bottom and get rid of 0

    bin_labels = []

    for idx in range(nr_of_classes):
        val_bin = idx / (nr_of_classes - 1.0)
        bin_labels.append(val_bin)

    # ----------------------
    # Add white bin and color
    # ----------------------
    color_list_copy = copy.copy(color_list)
    bin_labels_copy = copy.copy(bin_labels)

    # Add zero color in color list if a min_plus map
    if color_zero != False:
        insert_pos = 1
        #insert_pos = int(len(bin_labels)/2) #Middle position where 0 is positioned in bins
        #insert_pos = 0
        color_list_copy.insert(insert_pos, color_zero)
        bin_labels_copy.insert(insert_pos, float(placeholder_zero_color))
    else:
        pass

    # Create the custom color map
    color_bin_match_list = []
    for lbl, color in zip(bin_labels_copy, color_list_copy):
        color_bin_match_list.append((lbl, color))

    if 0 in bins:
        pass
    elif min(bins) > 0:
        logging.info("TT")
    else:
        bins.insert(len(bins), 0) # Add zero at the end

    cmap = LinearSegmentedColormap.from_list(
        'mycmap',
        color_bin_match_list)

    #logging.info("cmap " + str(cmap))
    # Reclassify
    lad_geopanda_shp['reclassified'] = lad_geopanda_shp[field_to_plot].apply(
        func=bin_mapping,
        nr_of_classes=nr_of_classes,
        class_bins=bins,
        placeholder_zero_color=float(placeholder_zero_color))

    return lad_geopanda_shp, cmap

def plot_lad_national(
        lad_geopanda_shp,
        legend_unit,
        field_to_plot,
        fig_name_part,
        result_path,
        color_palette,
        color_prop=False,
        user_classification=False,
        color_list=False,
        color_zero=False,
        bins=[],
        file_type="pdf", #"png" pdf
        plotshow=False
    ):
    """Create plot of LADs and store to map file (PDF) and csv file

    Arguments
    ---------
    lad_geopanda_shp : dataframe
        Geopanda dataframe
    legend_unit : str
        Unit of values
    field_to_plot : str
        Field name to plot in map
    fig_name_part : str
        Additional string naming variable
    result_path : str
        Path to figure to plot
    user_classification : bool
        Criteria if user classification or not
    bins : list
        Classification boders
    color_palette : list
        List name of color scheme
    color_list : list
        User defined color scheme
    color_prop : str
        If whether sequential or qualitative colors

    Info
    ----

        color_palette

            Colorbrewer defined color schemes can be found here:
            https://jiffyclub.github.io/palettable/colorbrewer/sequential/

        bins (Classification borders)

            If only positive numbers are to classify, the first entry must not be zero.
            If positive and negative numbers, a user defined  color schmene `color_list`
            generated with `colors_plus_minus_map` must be provided as an input and
            user_classification set to `user_classification`.

        Geopanda info:
        http://darribas.org/gds_scipy16/ipynb_md/02_geovisualization.html
        https://stackoverflow.com/questions/41783090/plotting-a-choropleth
        -map-with-geopandas-using-a-user-defined-classification-s
    """
    # ---------------------
    # Write results to csv
    # ---------------------
    out_data_path = os.path.join(
        result_path,
        "{}_{}.{}".format(fig_name_part, field_to_plot, 'csv'))

    write_data.write_result_txt(
        out_data_path,
        lad_geopanda_shp['name'],
        lad_geopanda_shp[field_to_plot])

    # ---------------------
    # Create figure
    # ---------------------
    fig_name = os.path.join(
        result_path,
        "{}_{}.{}".format(
            fig_name_part, field_to_plot, file_type))

    fig_map, axes = plt.subplots(
        1, figsize=(5, 8))

    legend_title = "unit [{}]".format(legend_unit)

    # --------------------
    # Plot polygon borders and set per default to white
    # --------------------
    lad_geopanda_shp.plot(
        ax=axes,
        linewidth=0.6,
        color=None,
        edgecolor='black')

    # -----------------------------
    # Own classification (work around)
    # -----------------------------
    if user_classification:

        # Color to assing zero values
        placeholder_zero_color = 0.00001

        # Get maximum and minum values
        rounding_digits = 10
        min_value = round(lad_geopanda_shp[field_to_plot].min(), rounding_digits)
        max_value = round(lad_geopanda_shp[field_to_plot].max(), rounding_digits)

        # Add maximum value
        logging.info(" {} {}".format(min_value, max_value))
        logging.info("FINAL BIN before" + str(bins))

        ###logging.info("FINAL BIN " + str(bins))
        lad_geopanda_shp_reclass, cmap = user_defined_classification(
            bins,
            min_value,
            max_value,
            lad_geopanda_shp,
            field_to_plot,
            legend_title,
            color_palette=color_palette,
            color_prop=color_prop,
            placeholder_zero_color=placeholder_zero_color,
            color_zero=color_zero,
            color_list=color_list)

        # Plot reclassified
        lad_geopanda_shp_reclass.plot(
            ax=axes,
            column='reclassified',
            legend=False,
            cmap=cmap,
            alpha=1,
            vmin=0,
            vmax=1)

        # ---------
        # Select all polygons with value 0 of attribute to plot
        # ---------
        all_zero_polygons = getattr(lad_geopanda_shp_reclass, 'reclassified')
        lad_geopanda_shp_zeros = lad_geopanda_shp_reclass[(all_zero_polygons == placeholder_zero_color)]

        # If more than 0 polygons are selected with classified number of zero, plot them
        if lad_geopanda_shp_zeros.shape[0] > 0:
            lad_geopanda_shp_zeros.plot(
                ax=axes,
                color=color_zero)
    else:
        logging.info("not user classification")

        # ----------------------------
        # Plot map with all value hues
        # -----------------------------
        # Creates hues values
        lad_geopanda_shp.plot(
            ax=axes,
            column=field_to_plot,
            legend=True)
        '''
        lad_geopanda_shp.plot(
            ax=axes,
            column=field_to_plot,
            cmap='OrRd',
            legend=True)
        '''

        # -----------------------------
        # Plot map with quantiles
        # -----------------------------
        '''
        lad_geopanda_shp.plot(
            axes=axes,
            column=field_to_plot,
            scheme='equal_interval', #quantiles'
            k=10,
            cmap='OrRd',
            legend=True)
        '''

        # -----------------------------
        # Plot map wtih quantiles
        # -----------------------------
        '''lad_geopanda_shp.plot(
            axes=axes,
            column=field_to_plot,
            scheme='equal_interval',
            k=5,
            cmap='OrRd',
            legend=True)'''

        #from pysal.esda.mapclassify import User_Defined
        #lad_geopanda_shp.User_Defined(lad_geopanda_shp, bins)
        '''lad_geopanda_shp.assign(cl=interval_data.yb).plot(
            column='cl',
            categorical=True,
            k=10,
            cmap='OrRd',
            linewidth=0.1,
            axes=axes,
            edgecolor='white',
            legend=True,
            scheme='User_Defined')'''

        # Legend taken form axis
        '''
        legend = axes.get_legend()
        legend.set_title(legend_title, prop={
            'family': 'arial',
            'size': 10})
        legend.set_frame_on(False)
        legend.set_bbox_to_anchor((0.5, -0.05))
        '''

    # Title
    fig_map.suptitle(field_to_plot + fig_name_part)

    # Make that not distorted
    plt.axis('equal')

    # Tight layout
    plt.margins(x=0)

    # Add space for legend
    plt.subplots_adjust(bottom=0.4)

    # Save figure
    plt.savefig(fig_name)

    # Show figure
    if plotshow:
        plt.show()

    plt.close()

def merge_data_to_shp(shp_gdp, merge_data, unique_merge_id):
    """Merge data to geopanda dataframe which is read from shapefile

    Arguments
    ----------
    shp_gdp : dataframe
        Geopanda dataframe from shapefile
    merge_data : dict
        Data to merge
    unique_merge_id : str
        Unique ID to make attribute merge

    Returns
    --------
    shp_gdp_merged : dataframe
        Geopanda containing merged dataframe

    Example for merge_data
    ------------------------
        merge_data = {
            name_of_new_data_column: [9999, 9999],
            merge_unique_ID: ['W06000016', 'S12000013']}

    More info: http://geopandas.org/mergingdata.html
    """
    # Create a GeoDataFrame with joing attribute and values
    merge_dataframe = pd.DataFrame(
        data=merge_data)

    shp_gdp_merged = shp_gdp.merge(
        merge_dataframe,
        on=unique_merge_id)

    return shp_gdp_merged

def plot_spatial_mapping_example(
        diffusion_vals,
        global_value,
        paths,
        regions,
        path_shapefile_input,
        plotshow=False
    ):
    """Figure plot

    Plot a map which distributes a global parameter to
    regional parameters based on diffusion values.

    """
    # Read LAD shapefile and create geopanda
    lad_geopanda_shp = gpd.read_file(path_shapefile_input)

    # Attribute merge unique Key
    unique_merge_id = 'name' #'geo_code'

    field_name = 'Fig_XX_spatial_diffusion_vals'

    # Calculate regional values
    regional_vals = {}
    for region in regions:
        regional_vals[region] = global_value * diffusion_vals[region]

    # Both need to be lists
    merge_data = {
        str(field_name): list(regional_vals.values()),
        str(unique_merge_id): list(regions)}

    # Merge to shapefile
    lad_geopanda_shp = merge_data_to_shp(
        lad_geopanda_shp,
        merge_data,
        unique_merge_id)

    # If user classified, defined bins  [x for x in range(0, 1000000, 200000)]
    #bins = [-4, -2, 0, 2, 4] # must be of uneven length containing zero if minus values
    #bins = [-15, -10, -5, 0, 5, 10, 15] 
    bins = [20, 30, 40, 50, 60, 70, 80, 90]

    if max(bins) < max(list(regional_vals.values())):
        raise Exception("Wrong bin definition: max_val: {}  min_val: {}".format(
            max(list(regional_vals.values())), min(list(regional_vals.values()))))

    color_palette = 'YlGn_9'# YlGn_9
    color_list, color_prop, user_classification, color_zero = colors_plus_minus_map(
        bins=bins,
        color_prop='qualitative',
        color_order=True,
        color_zero='#ffffff',
        color_palette=color_palette)

    plot_lad_national(
        lad_geopanda_shp=lad_geopanda_shp,
        legend_unit="%",
        field_to_plot=field_name,
        fig_name_part="lf_max_y",
        result_path=paths['data_results_PDF'],
        color_palette=color_palette,
        color_prop=color_prop,
        user_classification=user_classification,
        color_list=color_list,
        color_zero=color_zero,
        bins=bins,
        plotshow=plotshow)

    return

def spatial_maps(
        data,
        results_container,
        path_data_results_shapefiles,
        regions,
        fueltypes_nr,
        fueltypes,
        path_shapefile_input,
        plot_crit_dict,
        base_yr
    ):
    """Create map related files from results

    Arguments
    ---------
    results_container : dict
        Data container
    paths : dict
        Paths
    regions : list
        Region in a list with order how they are stored in result array
    fueltypes_nr : int
        Number of fueltypes
    """
    logging.info("... create spatial maps of results")

    # --------
    # Read LAD shapefile and create geopanda
    # --------

    # Single scenario run
    lad_geopanda_shp = gpd.read_file(path_shapefile_input)

    # Attribute merge unique Key
    unique_merge_id = 'name' #'geo_code'

    fueltype_str_to_create_maps = [
        'electricity'] #,
        #'gas']

    # ======================================
    # Peak max h all enduses (abs)
    # ======================================
    if plot_crit_dict['plot_abs_peak_h']:
        for year in results_container['ed_fueltype_regs_yh'].keys():
            
            for fueltype_str in fueltype_str_to_create_maps:

                
                fueltype = tech_related.get_fueltype_int(fueltype_str)
    
                # Calculate peak h across all regions
                field_name = 'peak_abs_h_{}_{}'.format(year, fueltype_str)

                # Get maxium demand of 8760h for every region
                h_max_gwh_regs = np.max(results_container['ed_fueltype_regs_yh'][year][fueltype], axis=1)
                #print("TOTAL peak fuel across all regs {} {} ".format(np.sum(h_max_gwh_regs), fueltype_str))

                data_to_plot = basic_functions.array_to_dict(h_max_gwh_regs, regions)

                # Both need to be lists
                merge_data = {
                    str(field_name): list(data_to_plot.values()),
                    str(unique_merge_id): list(regions)}

                # Merge to shapefile
                lad_geopanda_shp = merge_data_to_shp(
                    lad_geopanda_shp,
                    merge_data,
                    unique_merge_id)

                bins = [0.2, 0.4, 0.6, 0.8, 1.0, 1.2]

                color_list, color_prop, user_classification, color_zero = colors_plus_minus_map(
                    bins=bins,
                    color_prop='qualitative',
                    color_order=True,
                    color_zero='#ffffff',
                    color_palette='YlGnBu_7') #YlGnBu_9 #8a2be2 'YlGnBu_9'  'PuBu_8'

                # If user classified, defined bins
                plot_lad_national(
                    lad_geopanda_shp=lad_geopanda_shp,
                    legend_unit="GWh",
                    field_to_plot=field_name,
                    fig_name_part=field_name,
                    result_path=path_data_results_shapefiles,
                    color_palette='Dark2_7',
                    color_prop=color_prop,
                    user_classification=user_classification,
                    color_list=color_list,
                    bins=bins,
                    color_zero=color_zero)

    # ======================================
    # Peak max h all enduses (diff p)
    # ======================================
    if plot_crit_dict['plot_diff_peak_h']:
        for year in results_container['ed_fueltype_regs_yh'].keys():
            if year == base_yr:
                pass
            else:
                
                for fueltype_str in fueltype_str_to_create_maps:

                    
                    fueltype = tech_related.get_fueltype_int(fueltype_str)

                    # If total sum is zero, skip
                    if np.sum(results_container['ed_fueltype_regs_yh'][base_yr][fueltype]) == 0:
                        continue

                    ##if np.isnan(np.sum(results_container['ed_fueltype_regs_yh'][base_yr][fueltype])):
                    #    logging.info("Error: Contains nan entry {} {}".format(year, fueltype))
                    #    continue
                    #logging.info("============ {}  {}".format(fueltype, np.isnan(np.sum(results_container['ed_fueltype_regs_yh'][base_yr][fueltype]))))
                    #logging.info(results_container['ed_fueltype_regs_yh'][base_yr][fueltype])

                    # Calculate peak h across all regions
                    field_name = 'peak_diff_p_peak_h_{}_{}'.format(year, fueltype_str)

                    # Get maxium demand of 8760h for every region for base year
                    h_max_gwh_regs_by = np.max(results_container['ed_fueltype_regs_yh'][base_yr][fueltype], axis=1)

                    # Get maxium demand of 8760h for every region for current year
                    h_max_gwh_regs_cy = np.max(results_container['ed_fueltype_regs_yh'][year][fueltype], axis=1)
                    #print("TOTAL peak fuel across all regs {} {} ".format(np.sum(h_max_gwh_regs_cy), fueltype_str))

                    # Calculate difference in decimal
                    diff_p_h_max_regs = ((100 / h_max_gwh_regs_by) * h_max_gwh_regs_cy) - 100

                    data_to_plot = basic_functions.array_to_dict(diff_p_h_max_regs, regions)

                    merge_data = {
                        str(field_name): list(data_to_plot.values()),
                        str(unique_merge_id): list(regions)}

                    # Merge to shapefile
                    lad_geopanda_shp = merge_data_to_shp(
                        lad_geopanda_shp,
                        merge_data,
                        unique_merge_id)

                    '''bins_increments = 25
                    bins = get_reasonable_bin_values(
                        data_to_plot=list(data_to_plot.values()),
                        increments=bins_increments)'''
                    #bins = [10, 20, 40, 80, 160, 320, 640]
                    bins = [-160, -320, -640, -80, -40, -20, -10 ,0, 10, 20, 40, 80, 160, 320, 640]

                    color_list, color_prop, user_classification, color_zero = colors_plus_minus_map(
                        bins=bins,
                        color_prop='qualitative',
                        color_order=True,
                        color_zero='#ffffff',
                        color_palette='YlGnBu_7') #YlGnBu_9 #8a2be2 'PuBu_8'
   
                    plot_lad_national(
                        lad_geopanda_shp=lad_geopanda_shp,
                        legend_unit="GWh",
                        field_to_plot=field_name,
                        fig_name_part=field_name,
                        result_path=path_data_results_shapefiles,
                        color_palette='Dark2_7',
                        color_prop=color_prop,
                        user_classification=user_classification,
                        color_list=color_list,
                        bins=bins,
                        color_zero=color_zero)

    # ======================================
    # Load factors (absolute)
    # ======================================
    if plot_crit_dict['plot_load_factors']:
        for year in results_container['reg_load_factor_y'].keys():
            
            for fueltype_str in fueltype_str_to_create_maps:
                
                fueltype = tech_related.get_fueltype_int(fueltype_str)

                field_name = 'lf_{}_{}'.format(year, fueltype_str)

                results = basic_functions.array_to_dict(
                    results_container['reg_load_factor_y'][year][fueltype], regions)

                # Both need to be lists
                merge_data = {
                    str(field_name): list(results.values()),
                    str(unique_merge_id): list(regions)}

                # Merge to shapefile
                lad_geopanda_shp = merge_data_to_shp(
                    lad_geopanda_shp,
                    merge_data,
                    unique_merge_id)

                # ABSOLUTE
                bins = [40, 45, 50, 55, 60, 75, 80]
                #bins = [55, 60, 65, 70] # must be of uneven length containing zero
                color_list, color_prop, user_classification, color_zero = colors_plus_minus_map(
                    bins=bins,
                    color_prop='qualitative',
                    color_order=True,
                    color_zero='#ffffff',
                    color_palette='YlGnBu_9') #8a2be2 'YlGnBu_9'  'PuBu_8'

                # If user classified, defined bins
                plot_lad_national(
                    lad_geopanda_shp=lad_geopanda_shp,
                    legend_unit="%",
                    field_to_plot=field_name,
                    fig_name_part="lf_max_y",
                    result_path=path_data_results_shapefiles,
                    color_palette='Dark2_7',
                    color_prop=color_prop,
                    user_classification=user_classification,
                    color_list=color_list,
                    bins=bins,
                    color_zero=color_zero)

    # ======================================
    # Load factors (difference p)
    # ======================================
    if plot_crit_dict['plot_load_factors_p']:
        sim_yrs = list(results_container['reg_load_factor_y'].keys())

        final_yr = sim_yrs[-1]
        base_yr = sim_yrs[0]

        
        for fueltype_str in fueltype_str_to_create_maps:
            fueltype = tech_related.get_fueltype_int(fueltype_str)
            

            field_name = 'lf_diff_{}-{}_{}'.format(base_yr, final_yr, fueltype_str)

            lf_end_yr = basic_functions.array_to_dict(
                results_container['reg_load_factor_y'][final_yr][fueltype],
                regions)

            lf_base_yr = basic_functions.array_to_dict(
                results_container['reg_load_factor_y'][base_yr][fueltype],
                regions)

            # Calculate load factor difference base and final year (100 = 100%)
            diff_lf = {}
            for reg in regions:
                diff_lf[reg] = lf_end_yr[reg] - lf_base_yr[reg]

            # Both need to be lists
            merge_data = {
                str(field_name): list(diff_lf.values()),
                str(unique_merge_id): list(regions)}

            # Merge to shapefile
            lad_geopanda_shp = merge_data_to_shp(
                lad_geopanda_shp,
                merge_data,
                unique_merge_id)

            # If user classified, defined bins  [x for x in range(0, 1000000, 200000)]
            #bins = [-4, -2, 0, 2, 4] # must be of uneven length containing zero
            bins = [-15, -10, -5, 0, 5, 10, 15] # must be of uneven length containing zero

            color_list, color_prop, user_classification, color_zero = colors_plus_minus_map(
                bins=bins,
                color_prop='qualitative',
                color_order=True,
                color_zero='#ffffff',
                color_palette='Purples_9')

            plot_lad_national(
                lad_geopanda_shp=lad_geopanda_shp,
                legend_unit="%",
                field_to_plot=field_name,
                fig_name_part="lf_max_y",
                result_path=path_data_results_shapefiles,
                color_palette='Purples_9',
                color_prop=color_prop,
                user_classification=user_classification,
                color_list=color_list,
                color_zero=color_zero,
                bins=bins)

    # ======================================
    # Population
    # ======================================
    if plot_crit_dict['plot_population']:
        for year in results_container['ed_fueltype_regs_yh'].keys():

            field_name = 'pop_{}'.format(year)

            # Both need to be lists
            pop_data = data['scenario_data']['population'][year]

            merge_data = {
                #str(field_name): data['scenario_data']['population'][year].flatten().tolist(),
                str(field_name): list(data['scenario_data']['population'][year]),
                str(unique_merge_id): list(regions)}

            # Merge to shapefile
            lad_geopanda_shp = merge_data_to_shp(
                lad_geopanda_shp,
                merge_data,
                unique_merge_id)

            # If user classified, defined bins
            bins = [50000, 300000]

            plot_lad_national(
                lad_geopanda_shp=lad_geopanda_shp,
                legend_unit="people",
                field_to_plot=field_name,
                fig_name_part="pop_",
                result_path=path_data_results_shapefiles,
                color_palette='Dark2_7',
                color_prop='qualitative',
                user_classification=True,
                bins=bins)

    # ======================================
    # Total fuel (y) all enduses
    # ======================================
    for year in results_container['ed_fueltype_regs_yh'].keys():
        
        for fueltype_str in fueltype_str_to_create_maps:
            fueltype = tech_related.get_fueltype_int(fueltype_str)

            if plot_crit_dict['plot_total_demand_fueltype']:

                # ---------
                # Sum per enduse and year (y)
                # ---------
                field_name = 'y_{}_{}'.format(year, fueltype_str)

                # Calculate yearly sum across all regions
                yearly_sum_gwh = np.sum(
                    results_container['ed_fueltype_regs_yh'][year][fueltype],
                    axis=1)

                data_to_plot = basic_functions.array_to_dict(yearly_sum_gwh, regions)

                # Both need to be lists
                merge_data = {
                    str(field_name): list(data_to_plot.values()),
                    str(unique_merge_id): list(regions)}

                # Merge to shapefile
                lad_geopanda_shp = merge_data_to_shp(
                    lad_geopanda_shp,
                    merge_data,
                    unique_merge_id)

                # If user classified, defined bins
                plot_lad_national(
                    lad_geopanda_shp=lad_geopanda_shp,
                    legend_unit="GWh",
                    field_to_plot=field_name,
                    fig_name_part="_tot_all_enduses_y",
                    result_path=path_data_results_shapefiles,
                    color_palette='Dark2_7',
                    color_prop='qualitative',
                    user_classification=False)

            # ===============================================
            # Differences in percent per enduse and year (y)
            # ===============================================
            if plot_crit_dict['plot_differences_p'] and year > base_yr:

                field_name = 'y_diff_p_{}-{}_{}'.format(
                    base_yr, year, fueltype_str)

                # Calculate yearly sums
                yearly_sum_gwh_by = np.sum(
                    results_container['ed_fueltype_regs_yh'][base_yr][fueltype],
                    axis=1)

                yearly_sum_gwh_cy = np.sum(
                    results_container['ed_fueltype_regs_yh'][year][fueltype],
                    axis=1)

                # Calculate percentual difference
                p_diff = ((yearly_sum_gwh_cy / yearly_sum_gwh_by) * 100) - 100

                data_to_plot = basic_functions.array_to_dict(p_diff, regions)

                # Both need to be lists
                merge_data = {
                    str(field_name): list(data_to_plot.values()),
                    str(unique_merge_id): list(regions)}

                # Merge to shapefile
                lad_geopanda_shp = merge_data_to_shp(
                    lad_geopanda_shp,
                    merge_data,
                    unique_merge_id)

                # Test if nan vlaue in list and if yes, skipt this
                nan_entry = False
                nan_value = float('nan')
                for entry in list(data_to_plot.values()):
                    if math.isnan(entry):
                        nan_entry = True
                if nan_entry:
                    continue

                # ----
                # CAlculate classes for manual classification
                # ----                
                '''bins_increments = 25 #MAYBE NEEDS TO BE ADOPTED #TODO KAMEL
                bins = get_reasonable_bin_values(
                    data_to_plot=list(data_to_plot.values()),
                    increments=bins_increments)'''
                #bins = [10, 20, 40, 80, 160, 320, 640]
                bins = [-160, -320, -640, -80, -40, -20, -10 ,0, 10, 20, 40, 80, 160, 320, 640]

                color_list, color_prop, user_classification, color_zero = colors_plus_minus_map(
                    bins=bins,
                    color_prop='qualitative',
                    color_order=True)

                # Plot difference in % per fueltype of total fuel (y)
                plot_lad_national(
                    lad_geopanda_shp=lad_geopanda_shp,
                    legend_unit="GWh",
                    field_to_plot=field_name,
                    fig_name_part="tot_all_enduses_y_",
                    result_path=path_data_results_shapefiles,
                    color_palette='Purples_9',
                    color_prop=color_prop,
                    user_classification=user_classification,
                    color_list=color_list,
                    color_zero=color_zero,
                    bins=bins)

                # PLot not user classificaiton
                plot_lad_national(
                    lad_geopanda_shp=lad_geopanda_shp,
                    legend_unit="GWh",
                    field_to_plot=field_name,
                    fig_name_part="_huevalues_",
                    result_path=path_data_results_shapefiles,
                    color_palette='Purples_9',
                    color_prop=color_prop,
                    user_classification=False,
                    color_list=color_list,
                    color_zero=color_zero,
                    bins=bins)

def colors_plus_minus_map(
        bins,
        color_prop,
        color_order=True,
        color_zero='#ffffff',
        color_palette='Purples_9'
    ):
    """Create color scheme in case plus and minus classes
    are defined (i.e. negative and positive values to
    classify)

    Arguments
    ---------
    bins : list
        List with borders
    color_prop : str
        Type of color is not plus_minus map
    color_order : bool
        Criteria to switch colors
    user_classification : bool
        Criteria whether used classification or not
    color_zero : str, default=white hex color
        Color of zero values

    Returns
    -------
    color_list : list
        List with colors
    color_prop : str
        Type of colors
    user_classification : bool
        Wheter user defined color scheme or not
    color_zero : hex_color string
        Color of zero values
    """
    if bins == []:
        # regular classification
        return [], color_prop, False, False

    elif min(bins) < 0:
        #logging.info("negative bins" + str(bins))
        #print("negative bins" + str(bins))
        #if len(bins) > 10:
        #    raise Exception("Too many bins defined: Change interval criteria")
        #if len(bins) == 10 or len(bins) == 9: # add extra color to reach 10 colors
        if 1== 1:
            if color_order:
                color_list_pos = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors
                color_list_pos.append('#330808')
                color_list_neg = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors
                color_list_neg.append('#0e2814')
            else:
                color_list_neg = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors   # 'Reds_9'
                color_list_pos.append('#330808')
                color_list_pos = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors # 'Greens_9'
                color_list_pos.append('#0e2814')
        #elif len(bins) == 9:
        #    if color_order:
        #        color_list_pos = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors
        #        color_list_neg = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors
        #    else:
        #        color_list_neg = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors   # 'Reds_9'
        #        color_list_pos = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors # 'Greens_9'
        elif len(bins) == 8:
            if color_order:
                color_list_pos = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors
                color_list_neg = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors
            else:
                color_list_neg = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors   # 'Reds_9'
                color_list_pos = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors # 'Greens_9'
        else:
            # Colors pos and neg
            if color_order:
                color_list_pos = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors   # 'Reds_9'
                color_list_neg = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors # 'Greens_9'
            else:
                color_list_neg = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors   # 'Reds_9'
                color_list_pos = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors # 'Greens_9'

        # Number of categories
        #nr_of_cat_pos_neg = int((len(bins) -1) / 2)

        # Invert negative colors
        color_list_neg = color_list_neg[::-1]
        #logging.info("nr_of_cat_pos_neg " + str(nr_of_cat_pos_neg))

        color_list = []

        if max(bins) == 0:
            nr_of_cat_neg = int(len(bins)-1)
            for i in range(nr_of_cat_neg + 1): #add one to get class up to zero
                color_list.append(color_list_neg[i])
        if max(bins) <= 0:
            nr_of_cat_neg = int(len(bins))
            for i in range(nr_of_cat_neg + 1): #add one to get class up to zero
                color_list.append(color_list_neg[i])

            color_list.append(color_zero) # Add 0 color
        elif min(bins) >= 0:
            nr_of_cat_pos = int(len(bins))
            for i in range(nr_of_cat_pos + 1): #add one to get class up to zero
                color_list.append(color_list_neg[i])

            color_list.insert(0, color_zero) # Add 0 color
        else:
            nr_of_cat_neg = 0
            nr_of_cat_pos = 0
        
            for i in bins:
                if i < 0:
                    nr_of_cat_neg += 1
                elif i > 0:
                    nr_of_cat_pos += 1
                else:
                    pass

            for i in range(nr_of_cat_neg + 1): #add one to get class before first bin
                color_list.append(color_list_neg[i])

            for i in range(nr_of_cat_pos + 1): # add one to get class beyond last bin
                color_list.append(color_list_pos[i])

        return color_list, 'user_defined', True, color_zero

    elif min(bins) > 0:

        color_list = []

        if len(bins) == 10 or len(bins) == 9: # add extra color to reach 10 colors

            if color_order:
                if max(bins) < 0:
                    color_list_pos = getattr(palettable.colorbrewer.sequential, 'Greens_8').hex_colors
                    color_list_pos.append('#0e2814')
                else:
                    color_list_pos = getattr(palettable.colorbrewer.sequential, 'Reds_8').hex_colors
                    color_list_pos.append('#330808')
            else:
                if max(bins) < 0:
                    color_list_pos = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors # 'Greens_9
                    color_list_pos.append('#0e2814')
                else:
                    color_list_pos = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors   # 'Reds_9'
                    color_list_pos.append('#330808')
        else:
            if color_order:
                color_list_pos = getattr(palettable.colorbrewer.sequential, 'Reds_9').hex_colors   # 'Reds_9'
            else:
                color_list_pos = getattr(palettable.colorbrewer.sequential, 'Greens_9').hex_colors # 'Greens_9'
        #color_list_pos = getattr(palettable.colorbrewer.sequential, color_palette).hex_colors

        # Number of categories
        nr_of_cat_pos_neg = int((len(bins)))

        for i in range(nr_of_cat_pos_neg + 1): # add one to get class beyond last bin
            color_list.append(color_list_pos[i])

        return color_list, 'user_defined', True, color_zero

    else:
        # regular classification
        return [], color_prop, False, False



"""Functions related to technologies
"""
import logging
import numpy as np
from energy_demand.technologies import diffusion_technologies as diffusion
from energy_demand.read_write import read_data
from energy_demand.basic import lookup_tables

def test_if_tech_defined(enduse_fueltypes_techs):
    """Test if a technology has been configured,
    i.e. a fuel share has been assgined to one of the
    fueltpyes in `fuel_shares`.

    Arguments
    ---------
    enduse_fueltypes_techs : dict
        Configured technologies and fuel shares of an enduse

    Returns
    -------
    c_tech_defined : bool
        Criteria whether technologies have been configured
        for an enduse or not
    """
    c_tech_defined = False

    for fueltype in enduse_fueltypes_techs:
        if enduse_fueltypes_techs[fueltype] == {}:
            pass
        else:
            c_tech_defined = True
            break

    return c_tech_defined

def insert_placholder_techs(
        technologies,
        tech_p_by,
        specified_tech_enduse_by
    ):
    """If no technology is defined for a fueltype in an enduse add
    a dumppy technology. This is necessary because the model needs
    a technology for every fueltype in an enduse.

    Arguments
    ----------
    technologies : dict
        Technologies
    tech_p_by : dict
        Fuel assignement of technologies in base year
    specified_tech_enduse_by : dict
        Technologies per enduse

    Returns
    -------
    tech_p_by : dict
        Fuel assignement of technologies in base year
    specified_tech_enduse_by : dict
        Technologies per enduse
    technologies : dict
        Technologies
    """
    out_tech_p_by = {}
    for enduse, enduse_fueltypes_techs in tech_p_by.items():
        out_tech_p_by[enduse] = {}
        if list(enduse_fueltypes_techs.keys())[0] != 0:

            for sector in enduse_fueltypes_techs:
                out_tech_p_by[enduse][sector] = {}

                # Test if a technology is defined in any fueltype
                c_tech_defined = test_if_tech_defined(enduse_fueltypes_techs[sector])

                if not c_tech_defined:
                    for fueltype in enduse_fueltypes_techs[sector]:
                        if enduse_fueltypes_techs[sector][fueltype] == {}:

                            # Assign total fuel demand to dummy technology
                            out_tech_p_by[enduse][sector][fueltype] = {"placeholder_tech": 1.0}

                    specified_tech_enduse_by[enduse][sector].append("placeholder_tech")
                else:
                    out_tech_p_by[enduse][sector] = enduse_fueltypes_techs[sector]
        else:
            # If no sector is defined, add a dummy None setor
            dummy_sector = None

            out_tech_p_by[enduse][dummy_sector] = {}
            c_tech_defined = test_if_tech_defined(enduse_fueltypes_techs)
            if not c_tech_defined:
                for fueltype in enduse_fueltypes_techs:
                    if enduse_fueltypes_techs[fueltype] == {}:
                        out_tech_p_by[enduse][dummy_sector][fueltype] = {"placeholder_tech": 1.0}

                specified_tech_enduse_by[enduse][dummy_sector].append("placeholder_tech")
            else:
                out_tech_p_by[enduse][dummy_sector] = enduse_fueltypes_techs

    # Insert placeholder technology
    technologies['placeholder_tech'] = read_data.TechnologyData(
        eff_by=1,
        eff_ey=1,
        year_eff_ey=2100,
        eff_achieved=1,
        diff_method='linear',
        tech_type='placeholder_tech')

    return out_tech_p_by, specified_tech_enduse_by, technologies

def calc_hp_eff(
        temp_yh,
        efficiency_intersect,
        t_base_heating
    ):
    """Calculate efficiency of heat pumps according to
    temperature difference.

    Arguments
    ----------
    temp_yh : array
        Temperatures for every hour in a year
    efficiency_intersect : float
        Y-value (Efficiency) at 10 degree difference
    t_base_heating : float
        Base temperature for heating

    Return
    ------
    av_eff_hp : float
        Average eEfficiency for every hour in a year

    Note
    -----
    - For every hour the temperature difference is calculated and the efficiency
      of the heat pump calculated based on efficiency assumptions

    - The efficiency assumptions of the heat pump are taken from
      Staffell et al. (2012). Expexted average efficiencies for
      heating are expected to be for ASHP around 3.3 - 3.9, for
      GSHP around 4.5 - 5.4.
      For hot water, it is around 2.3 - 2.8 for ASHP and
      3.1-3.8 for GSHP.

      Staffell, I., Brett, D., Brandon, N., & Hawkes, A. (2012). A review of domestic heat pumps.
      Energy & Environmental Science, 5(11), 9291. https://doi.org/10.1039/c2ee22653g
    """
    # Calculate temperature difference to t_base_heating
    temp_difference_temp_yh = t_base_heating - temp_yh

    #Ignore all hours where no heating is necessary
    temp_difference_temp_yh[temp_difference_temp_yh < 0] = 0

    # Calculate average efficiency of heat pumps over full year
    av_eff_hp = eff_heat_pump(
        temp_difference_temp_yh,
        efficiency_intersect)

    return float(av_eff_hp)

def eff_heat_pump(temp_diff, efficiency_intersect, m_slope=-.08, h_diff=10):
    """Calculate efficiency of heat pumps

    Arguments
    ----------
    temp_diff: array
        Temperature difference between base temperature and temperature
    efficiency_intersect : float,default=-0.08
        Extrapolated intersect at temperature
        difference of 10 degree (which is treated as efficiency)
    m_slope : float, default=10
        Temperature dependency of heat pumps (slope)
    h_diff : float
        Temperature difference

    Return
    ------
    efficiency_hp_mean : array
        Mean efficiency of heat pump

    Note
    ----
    Because the efficieny of heat pumps is temperature dependent,
    the efficiency needs to be calculated based on slope and
    intersect which is provided as input for temp difference 10
    and treated as efficiency
    """
    efficiency_hp = m_slope * temp_diff + (efficiency_intersect - (m_slope * h_diff))

    # Calculate average efficiency over whole year
    efficiency_hp_mean = np.average(efficiency_hp)

    return efficiency_hp_mean

def get_fueltype_str(fueltype_lu, fueltype_nr):
    """Read from dict the fueltype string based on fueltype KeyError

    Arguments
    ---------
    fueltype : dict
        Fueltype lookup dictionary
    fueltype_nr : int
        Key which is to be found in lookup dict

    Returns
    -------
    fueltype_in_string : str
        Fueltype string
    """
    fueltype_lu = lookup_tables.basic_lookups()['fueltypes']

    for fueltype_str in fueltype_lu:
        if fueltype_lu[fueltype_str] == fueltype_nr:
            return fueltype_str

def get_fueltype_int(fueltype_str):
    """Read from dict the fueltype string based on fueltype KeyError

    Arguments
    ---------
    fueltype_str : int
        Key which is to be found in lookup dict

    Returns
    -------
    fueltype_in_string : str
        Fueltype string
    """
    fueltype_lu = lookup_tables.basic_lookups()['fueltypes']

    if not fueltype_str:
        return None
    else:
        return fueltype_lu[fueltype_str]

def calc_av_heat_pump_eff_ey(technologies, heat_pump_assump):
    """Calculate end year average efficiency of
    heat pumps depending on split of heat pumps

    Arguments
    ---------
    technologies : dict
        Technologies
    heat_pump_assump : dict
        Ditionary with split of heat pumps for every fueltype

    Return
    ------
    technologies : dict
        Updated technologies
    """
    for fueltype, heat_pump_split in heat_pump_assump.items():
        av_eff_hps_ey = 0

        for heat_pump_type, share_heat_pump in heat_pump_split.items():

            # End year efficiency of heat pump
            eff_heat_pump_ey = technologies[heat_pump_type].eff_ey

            # Calc average values
            av_eff_hps_ey += share_heat_pump * eff_heat_pump_ey

        # Add average 'av_heat_pumps' to technology dict
        name_av_hp = "heat_pumps_{}".format(fueltype)

        # Add new averaged technology
        technologies[name_av_hp].eff_ey = av_eff_hps_ey

    return technologies

def generate_heat_pump_from_split(technologies, heat_pump_assump, fueltypes):
    """Define average new heat pumptechnologies 'av_heat_pump_fueltype' with
    efficiency depending on installed ratio of heat pumps

    Arguments
    ----------
    technologies : dict
        Technologies
    heat_pump_assump : dict
        The split of the ASHP and GSHP
    fueltypes : dict
        Fueltypes lookup

    Returns
    -------
    technologies : dict
        Technologies with added averaged heat pump technologies for every fueltype
    temp_dependent_tech_list : list
        List with added temperature dependent technologies

    Note
    ----
    - Market Entry of different technologies must be the same year!
      (the lowest is selected if different years)
    - Diff method is linear
    """
    temp_dependent_tech_list = []
    heat_pumps = []

    # Calculate average efficiency of heat pump depending on installed ratio
    for fueltype in heat_pump_assump:
        av_eff_hps_by, av_eff_hps_ey, eff_achieved_av, market_entry_lowest = 0, 0, 0, 2200
        av_year_eff_ey = 0

        for heat_pump_type in heat_pump_assump[fueltype]:
            share_heat_pump = heat_pump_assump[fueltype][heat_pump_type]

            eff_heat_pump_by = technologies[heat_pump_type].eff_by
            eff_heat_pump_ey = technologies[heat_pump_type].eff_ey
            eff_achieved = technologies[heat_pump_type].eff_achieved
            market_entry = technologies[heat_pump_type].market_entry
            tech_max_share = technologies[heat_pump_type].tech_max_share

            av_year_eff_ey += technologies[heat_pump_type].year_eff_ey

            # Calc average values
            av_eff_hps_by += share_heat_pump * eff_heat_pump_by
            av_eff_hps_ey += share_heat_pump * eff_heat_pump_ey
            eff_achieved_av += share_heat_pump * eff_achieved

            if market_entry < market_entry_lowest:
                market_entry_lowest = market_entry

        # Calculate average year until efficiency improvements are implemented
        av_year_eff_ey = av_year_eff_ey / len(heat_pump_assump[fueltype])

        # Add average 'av_heat_pumps' to technology dict
        name_av_hp = "heat_pumps_{}".format(fueltype)

        # Add technology to temperature dependent technology list
        temp_dependent_tech_list.append(name_av_hp)

        # Add new averaged technology
        technologies[name_av_hp] = read_data.TechnologyData(
            fueltype=fueltype,
            eff_by=av_eff_hps_by,
            eff_ey=av_eff_hps_ey,
            year_eff_ey=av_year_eff_ey,
            eff_achieved=eff_achieved_av,
            diff_method='linear',
            market_entry=market_entry_lowest,
            tech_type='heating_non_const',
            tech_max_share=tech_max_share)

        heat_pumps.append(name_av_hp)

    return technologies, temp_dependent_tech_list, heat_pumps

def calc_eff_cy(
        base_yr,
        curr_yr,
        eff_by,
        eff_ey,
        yr_until_changed,
        f_eff_achieved,
        diff_method
    ):
    """Calculate efficiency of current year based on efficiency
    assumptions and achieved efficiency

    Arguments
    ----------
    base_yr : int
        Base year
    curr_yr : int
        Current year
    eff_by : dict
        Base year efficiency
    eff_ey : dict
        End year efficiency
    yr_until_changed : int
        Year for which the eff_ey is defined
    f_eff_achieved : dict
        Efficiency achievement factor (how much of the efficiency is achieved)
    diff_method : str
        Diffusion method

    Returns
    -------
    eff_cy : array
        Array with hourly efficiency of current year

    Notes
    -----
    The development of efficiency improvements over time is assumed to be linear
    This can however be changed with the `diff_method` attribute

    NICETOHAVE: Generate two types of sigmoid (convex & concav)
    """
    if diff_method == 'linear':
        # Theoretical maximum efficiency potential (linear improvement)
        theor_max_eff = diffusion.linear_diff(
            base_yr,
            curr_yr,
            eff_by,
            eff_ey,
            yr_until_changed)

        # Differencey in efficiency change
        max_eff_gain = theor_max_eff - eff_by

    elif diff_method == 'sigmoid':
        # Theoretical maximum efficiency potential (sigmoid improvement)
        diff_cy = diffusion.sigmoid_diffusion(
            base_yr,
            curr_yr,
            yr_until_changed,
            sig_midpoint=0,
            sig_steepness=1)

        # Differencey in efficiency change
        max_eff_gain = diff_cy * (eff_ey - eff_by)
    else:
        if not diff_method:
            return None
        else:
            logging.exception("Not correct diffusion assigned %s", diff_method)

    # Consider actual achieved efficiency
    actual_eff_gain = max_eff_gain * f_eff_achieved

    # Actual efficiency potential
    eff_cy = eff_by + actual_eff_gain

    return eff_cy

def generate_ashp_gshp_split(gshp_fraction):
    """Assing split for each fueltype of heat pump technologies,
    i.e. the share of GSHP of all heat pumps

    Arguments
    ----------
    gshp_fraction : float
        Fraction of GSHP (GSHP + ASHP = 100%)

    Returns
    --------
    installed_heat_pump_by : dict
        Ditionary with split of heat pumps for every fueltype
    """
    ashp_fraction = 1 - gshp_fraction

    installed_heat_pump_by = {
        'hydrogen': {
            'heat_pump_ASHP_hydrogen': ashp_fraction,
            'heat_pump_GSHP_hydrogen': gshp_fraction},
        'electricity': {
            'heat_pump_ASHP_electricity': ashp_fraction,
            'heat_pump_GSHP_electricity': gshp_fraction}
    }

    return installed_heat_pump_by

def get_technologies_per_fueltype(enduse_techs, technologies, fueltype):
    """Get technologies for fueltype
    """
    technologies_fueltype = []
    for tech in enduse_techs:
        tech_fueltype = technologies[tech].fueltype_str
        if tech_fueltype == fueltype:
            technologies_fueltype.append(tech)

    return technologies_fueltype



"""Read in model results and plot results
"""
import os
import pandas as pd
import numpy as np

from energy_demand.read_write import data_loader, read_data
from energy_demand.basic import date_prop
from energy_demand.basic import basic_functions
from energy_demand.basic import lookup_tables
from energy_demand.read_write import read_weather_results
from energy_demand.plotting import fig_3_weather_map
from energy_demand.technologies import tech_related
from energy_demand.plotting import fig_3_plot_over_time

def main(
        scenarios_path,
        path_shapefile_input,
        base_yr,
        simulation_yrs_to_plot
    ):
    """Read in all results and plot PDFs

    Arguments
    ----------
    scenarios_path : str
        Path to results
    path_shapefile_input : str
        Path to shapefile
    plot_crit_dict : dict
        Criteria to select plots to plot
    base_yr : int
        Base year
    comparison_year : int
        Year to generate comparison plots
    """
    print("Start creating plots")

    # -------------------
    # Create temperatuere figur plot
    # -------------------
    plot_weather_chart = False
    if plot_weather_chart:
        from energy_demand.plotting import fig3_weather_at_home_plot
        path_weather_data = "//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/J-MARIUS_data/_weather_realisation"
        fig3_weather_at_home_plot.plotting_weather_data(path_weather_data)

    # -------------------
    # Create result folder
    # -------------------
    result_path = os.path.join(scenarios_path, '_results_weather_plots')
    basic_functions.del_previous_setup(result_path)
    basic_functions.create_folder(result_path)

    x_chart_yrs_storage = {}

    for simulation_yr_to_plot in simulation_yrs_to_plot:
        print("=================")
        print("...simulation_yr_to_plot: " + str(simulation_yr_to_plot))
        print("=================")
        data = {}
        x_chart_yrs_storage[simulation_yr_to_plot] = {}

        # ---------------------------------------------------------
        # Iterate folders and read out all weather years and stations
        # ---------------------------------------------------------
        to_ignores = [
            'model_run_pop',
            'PDF_validation',
            '_results_weather_plots']

        endings_to_ignore = [
            '.pdf',
            '.txt',
            '.ini']

        all_scenarios_incl_ignored = os.listdir(scenarios_path)
        all_scenarios = []
        for scenario in all_scenarios_incl_ignored:
            if scenario not in to_ignores:
                all_scenarios.append(scenario)

        scenario_result_container = []
        for scenario_nr, scenario_name in enumerate(all_scenarios):
            print(" ")
            print("Scenario: {}".format(scenario_name), flush=True)
            print(" ")
            scenario_path = os.path.join(scenarios_path, scenario_name)
            all_result_folders = os.listdir(scenario_path)
            
            paths_folders_result = []

            for result_folder in all_result_folders:
                if result_folder not in to_ignores and result_folder[-4:] not in endings_to_ignore:
                    paths_folders_result.append(
                        os.path.join(scenario_path, result_folder))

            fueltype_str_to_create_maps = ['electricity']

            fueltype_str ='electricity'
            fueltype_elec_int = tech_related.get_fueltype_int('electricity')

            ####################################################################
            # Collect regional simulation data for every realisation
            ####################################################################
            total_regional_demand_electricity = pd.DataFrame()
            peak_hour_demand = pd.DataFrame()
            peak_hour_demand_per_person = pd.DataFrame()
            national_peak = pd.DataFrame()
            regional_share_national_peak = pd.DataFrame()
            regional_share_national_peak_pp = pd.DataFrame()
            national_electricity = pd.DataFrame()
            national_gas = pd.DataFrame()
            national_hydrogen = pd.DataFrame()
            national_heating_peak = pd.DataFrame()

            for path_result_folder in paths_folders_result:
                print("... path_result_folder: {}".format(path_result_folder), flush=True)
                data = {}
                ed_national_heating_peak = {}

                try:

                    # ================================
                    # Loading in only heating peak demands (seperate calculations)
                    # ================================
                    
                    # Simulation information is read in from .ini file for results
                    data['enduses'], data['assumptions'], data['regions'] = data_loader.load_ini_param(os.path.join(path_result_folder))
                    pop_data = read_data.read_scenaric_population_data(os.path.join(path_result_folder, 'model_run_pop'))
                    path_result_folder_heating = os.path.join(path_result_folder, 'simulation_results')
                    path_result_folder_model_runs = os.path.join(path_result_folder_heating, 'model_run_results_txt')
                    data['lookups'] = lookup_tables.basic_lookups()

                    # Read in heating deamnds
                    path_heating_demands = os.path.join(path_result_folder_model_runs, 'enduse_specific_results')
                    all_files = os.listdir(path_heating_demands)
                    for file_name in all_files:
                        
                        ending  = file_name[-4:]
                        if ending == ".npy":
                            year = int(file_name.split("__")[2][:-4])
                            file_path = os.path.join(path_heating_demands, file_name)
                            heating_demand = np.load(file_path)
                            maximum_hour_of_peak_day = heating_demand[fueltype_elec_int].argmax() #get maxim hour of peak day
                            ed_national_heating_peak[year] = heating_demand[fueltype_elec_int][maximum_hour_of_peak_day]

                    simulation_yrs_result = [ed_national_heating_peak[year] for year in simulation_yrs_to_plot]
                    realisation_data = pd.DataFrame(
                        [simulation_yrs_result],
                        columns=data['assumptions']['sim_yrs'])
                    national_heating_peak = national_heating_peak.append(realisation_data)

                except:
                    raise Exception("... no heating peak data available " + str(path_result_folder))

                try:
                    # Simulation information is read in from .ini file for results
                    data['enduses'], data['assumptions'], data['regions'] = data_loader.load_ini_param(os.path.join(path_result_folder))
                    pop_data = read_data.read_scenaric_population_data(os.path.join(path_result_folder, 'model_run_pop'))
                    path_result_folder = os.path.join(path_result_folder, 'simulation_results')
                    path_result_folder_model_runs = os.path.join(path_result_folder, 'model_run_results_txt')
                    data['lookups'] = lookup_tables.basic_lookups()

                    # Other information is read in
                    data['assumptions']['seasons'] = date_prop.get_season(year_to_model=2015)
                    data['assumptions']['model_yeardays_daytype'], data['assumptions']['yeardays_month'], data['assumptions']['yeardays_month_days'] = date_prop.get_yeardays_daytype(year_to_model=2015)

                    # --------------------------------------------
                    # Reading in results from different model runs
                    # --------------------------------------------
                    results_container = read_weather_results.read_in_weather_results(
                        path_result_folder_model_runs,
                        data['assumptions']['seasons'],
                        data['assumptions']['model_yeardays_daytype'],
                        pop_data,
                        fueltype_str='electricity')

                    # --Total demand (dataframe with row: realisation, column=region)
                    realisation_data = pd.DataFrame(
                        [results_container['ed_reg_tot_y'][simulation_yr_to_plot][fueltype_elec_int]],
                        columns=data['regions'])
                    total_regional_demand_electricity = total_regional_demand_electricity.append(realisation_data)

                    # National per fueltype electricity
                    simulation_yrs_result = [results_container['national_all_fueltypes'][year][fueltype_elec_int] for year in simulation_yrs_to_plot]

                    realisation_data = pd.DataFrame(
                        [simulation_yrs_result],
                        columns=data['assumptions']['sim_yrs'])
                    national_electricity = national_electricity.append(realisation_data)

                    # National per fueltype gas
                    fueltype_gas_int = tech_related.get_fueltype_int('gas')
                    simulation_yrs_result = [results_container['national_all_fueltypes'][year][fueltype_gas_int] for year in simulation_yrs_to_plot]

                    realisation_data = pd.DataFrame(
                        [simulation_yrs_result],
                        columns=data['assumptions']['sim_yrs'])
                    national_gas = national_gas.append(realisation_data)

                    # National per fueltype hydrogen
                    fueltype_hydrogen_int = tech_related.get_fueltype_int('hydrogen')
                    simulation_yrs_result = [results_container['national_all_fueltypes'][year][fueltype_hydrogen_int] for year in simulation_yrs_to_plot]

                    realisation_data = pd.DataFrame(
                        [simulation_yrs_result],
                        columns=data['assumptions']['sim_yrs'])
                    national_hydrogen = national_hydrogen.append(realisation_data)

                    # --Peak hour demand per region (dataframe with row: realisation, column=region)
                    realisation_data = pd.DataFrame(
                        [results_container['ed_reg_peakday_peak_hour'][simulation_yr_to_plot][fueltype_elec_int]],
                        columns=data['regions'])

                    peak_hour_demand = peak_hour_demand.append(realisation_data)

                    # --Peak hour demand / pop per region (dataframe with row: realisation, column=region)
                    realisation_data = pd.DataFrame(
                        [results_container['ed_reg_peakday_peak_hour_per_pop'][simulation_yr_to_plot][fueltype_elec_int]],
                        columns=data['regions'])

                    peak_hour_demand_per_person = peak_hour_demand_per_person.append(realisation_data)

                    # --National peak
                    simulation_yrs_result = [results_container['national_peak'][year][fueltype_elec_int] for year in simulation_yrs_to_plot]

                    realisation_data = pd.DataFrame(
                        [simulation_yrs_result],
                        columns=data['assumptions']['sim_yrs'])
                    national_peak = national_peak.append(realisation_data)

                    # --Regional percentage of national peak demand
                    realisation_data = pd.DataFrame(
                        [results_container['regional_share_national_peak'][simulation_yr_to_plot]],
                        columns=data['regions'])
                    regional_share_national_peak = regional_share_national_peak.append(realisation_data)
                
                    # --Regional percentage of national peak demand per person
                    realisation_data = pd.DataFrame(
                        [results_container['regional_share_national_peak_pp'][simulation_yr_to_plot]],
                        columns=data['regions'])

                    regional_share_national_peak_pp = regional_share_national_peak_pp.append(realisation_data)

                except:
                    raise Exception("The run '{}' is corrupted".format(path_result_folder), flush=True) 

            # Add to scenario container
            result_entry = {
                'national_heating_peak': national_heating_peak,
                'scenario_name': scenario_name,
                'peak_hour_demand': peak_hour_demand,
                'peak_hour_demand_per_person': peak_hour_demand_per_person,
                'national_peak': national_peak,
                'regional_share_national_peak': regional_share_national_peak,
                'regional_share_national_peak_pp': regional_share_national_peak_pp,
                'total_regional_demand_electricity': total_regional_demand_electricity,
                'national_electricity': national_electricity,
                'national_gas': national_gas,
                'national_hydrogen': national_hydrogen}

            scenario_result_container.append(result_entry)

            # ---------------------------------------------------------------
            # TEST PLOT X-axis: Contribution to peak y-axis: Std: deviation
            # ---------------------------------------------------------------
            x_chart_yrs_storage[simulation_yr_to_plot][scenario_name] = result_entry

        # ------------------------------
        # Plot national sum over time per fueltype and scenario
        # ------------------------------
        try:
            crit_smooth_line = True
            seperate_legend = True
            print("... plotting national sum of fueltype over time ")
            fig_3_plot_over_time.fueltypes_over_time(
                scenario_result_container=scenario_result_container,
                sim_yrs=data['assumptions']['sim_yrs'],
                fig_name="fueltypes_over_time__{}__{}.pdf".format(simulation_yr_to_plot, fueltype_str),
                fueltypes=['electricity', 'gas', 'hydrogen'],
                result_path=result_path,
                unit='TWh',
                plot_points=True,
                crit_smooth_line=crit_smooth_line,
                seperate_legend=seperate_legend)
        except:
            raise Exception("FAILS national sum")

        # ------------------------------
        # Plot national peak change over time for each scenario including weather variability
        # ------------------------------
        try:
            fig_3_plot_over_time.scenario_over_time(
                scenario_result_container=scenario_result_container,
                field_name='national_peak',
                sim_yrs=data['assumptions']['sim_yrs'],
                fig_name="scenarios_peak_over_time__{}__{}.pdf".format(simulation_yr_to_plot, fueltype_str),
                plot_points=True,
                result_path=result_path,
                crit_smooth_line=crit_smooth_line,
                seperate_legend=seperate_legend)
        except:
            raise Exception("FAILED")
            pass

        # ------------------------------
        # Plot heating peak change over time for each scenario including weather variability
        # ------------------------------
        try:
            fig_3_plot_over_time.scenario_over_time(
                scenario_result_container=scenario_result_container,
                field_name='national_heating_peak',
                sim_yrs=data['assumptions']['sim_yrs'],
                fig_name="scenarios_heating_peak_over_time__{}__{}.pdf".format(simulation_yr_to_plot, fueltype_str),
                plot_points=True,
                result_path=result_path,
                crit_smooth_line=crit_smooth_line,
                seperate_legend=seperate_legend)
        except:
            raise Exception("FAILED")
            pass

    ## ------------------------------
    ## Plotting x-chart
    ## ------------------------------
    fig_3_plot_over_time.plot_std_dev_vs_contribution(
        scenario_result_container=x_chart_yrs_storage,
        sim_yrs=data['assumptions']['sim_yrs'],
        fig_name="_scenarios_4_chart_absolute.pdf",
        fueltypes=['electricity'],
        result_path=result_path,
        path_shapefile_input=path_shapefile_input,
        unit='TWh',
        plot_points=True)

    print("===================================")
    print("... finished reading and plotting results")
    print("===================================")

# Code to run charts generation for weather paper
main(
    scenarios_path="//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/_p3_weather_final",
    path_shapefile_input="C:/Users/cenv0553/ED/data/region_definitions/lad_2016_uk_simplified.shp",
    base_yr=2015,
    simulation_yrs_to_plot=[2015, 2020, 2025, 2030, 2035, 2040, 2045, 2050])



"""
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from energy_demand.plotting import basic_plot_functions

def run(
        name_fig,
        path_result,
        ed_yh,
        days_to_plot=365,
        plot_crit=False
    ):
    """Plot individual enduse
    """
    ed_yh_365 = np.copy(ed_yh.reshape(365, 24))

    nr_of_h_to_plot = len(days_to_plot) * 24

    x_data = range(nr_of_h_to_plot)

    y_calculated = []
    for day in days_to_plot:
        for hour in range(24):
            y_calculated.append(ed_yh_365[day][hour])

    # ----------
    # Plot figure
    # ----------
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(16, 8))

    plt.plot(
        x_data,
        y_calculated,
        label='model',
        linestyle='--',
        linewidth=0.5,
        fillstyle='full',
        color='blue')

    plt.xlim([0, 8760])
    plt.margins(x=0)
    plt.axis('tight')

    # ----------
    # Labelling
    # ----------
    plt.xlabel("hour", fontsize=10)
    plt.ylabel("uk elec use [GW]", fontsize=10)

    plt.legend(frameon=False)

    plt.savefig(os.path.join(path_result, name_fig))

    if plot_crit:
        plt.show()
        plt.close()
    else:
        plt.close()



"""Plot stacked enduses for each submodel
"""
import numpy as np
import matplotlib.pyplot as plt
#from scipy.stats import mstats
import pandas as pd

from energy_demand.technologies import tech_related
from energy_demand.plotting import basic_plot_functions

def sum_all_enduses_fueltype(
        data_enduses,
        fueltype_str=False
    ):
    """Sum across all enduses and fueltypes
    TODO
    """
    y_values_enduse_yrs = {}
    for year in data_enduses.keys():
        for enduse in data_enduses[year].keys():

            # Sum across all fueltypes for every hour
            tot_across_fueltypes = data_enduses[year][enduse]

            try:
                y_values_enduse_yrs[year] += tot_across_fueltypes
            except KeyError:
                y_values_enduse_yrs[year] = tot_across_fueltypes

    return y_values_enduse_yrs

def run(
        data_input,
        simulation_yr_to_plot,
        period_h,
        fueltype_str,
        fig_name
    ):
    """

    https://stackoverflow.com/questions/18313322/plotting-quantiles-median-and-spread-using-scipy-and-matplotlib

    https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.quantile.html

    """
    # Select period and fueltype
    fueltype_int = tech_related.get_fueltype_int(fueltype_str)

    # -----------------------------------------------------------
    # Iterate overall weather_yrs and store data in dataframe
    # (columns = timestep, rows: value of year)
    # -----------------------------------------------------------
    columns = period_h # hours in 8760 hours

    # List of selected data for every weather year (which is then converted to array)
    weather_yrs_data = []

    print("Weather yrs: " + str(list(data_input.keys())), flush=True)

    for weather_yr, data_weather_yr in data_input.items():

        # Weather year specific data
        data_input_fueltype = data_weather_yr[simulation_yr_to_plot][fueltype_int]     # Select fueltype
        data_input_reshape = data_input_fueltype.reshape(8760)  # reshape
        data_input_selection_hrs = data_input_reshape[period_h] # select period
        weather_yrs_data.append(data_input_selection_hrs)

    weather_yrs_data = np.array(weather_yrs_data)

    # Create dataframe
    df = pd.DataFrame(weather_yrs_data, columns=columns)

    # Calculate quantiles
    quantile_95 = 0.95
    quantile_05 = 0.05

    df_q_95 = df.quantile(quantile_95)
    df_q_05 = df.quantile(quantile_05)

    #Transpose for plotting purposes
    df = df.T
    df_q_95 = df_q_95.T
    df_q_05 = df_q_05.T

    fig = plt.figure() #(figsize = cm2inch(10,10))

    ax = fig.add_subplot(111)

    # 2015 weather year
    data_2015 = data_weather_yr[2015][fueltype_int].reshape(8760)[period_h]

    # ---------------
    # Smoothing lines
    # ---------------
    try:
        period_h_smoothed, df_q_95_smoothed = basic_plot_functions.smooth_data(period_h, df_q_95, num=40000)
        period_h_smoothed, df_q_05_smoothed = basic_plot_functions.smooth_data(period_h, df_q_05, num=40000)
        period_h_smoothed, data_2015_smoothed = basic_plot_functions.smooth_data(period_h, data_2015, num=40000)
    except:
        period_h_smoothed = period_h
        df_q_95_smoothed = df_q_95
        df_q_05_smoothed = df_q_05
        data_2015_smoothed = data_2015

    plt.plot(period_h_smoothed, data_2015_smoothed, color='tomato', linestyle='-', linewidth=2, label="2015 weather_yr")
    plt.plot(period_h_smoothed, df_q_05_smoothed, color='black', linestyle='--', linewidth=0.5, label="0.05")
    plt.plot(period_h_smoothed, df_q_95_smoothed, color='black', linestyle='--', linewidth=0.5, label="0.95")

    # -----------------
    # Uncertainty range
    # -----------------
    plt.fill_between(
        period_h_smoothed, #x
        df_q_95_smoothed,  #y1
        df_q_05_smoothed,  #y2
        alpha=.25,
        facecolor="grey",
        label="uncertainty band")

    plt.legend(
        prop={
            'family':'arial',
            'size': 10},
        loc='best',
        frameon=False,
        shadow=True)

    plt.show()
    




"""Plot stacked enduses per sector
"""
import logging
import numpy as np
import matplotlib.pyplot as plt

from energy_demand.basic import conversions
from energy_demand.plotting import basic_plot_functions
##import matplotlib.colors as colors #for color_name in colors.cnmaes:

def run(
        lookups,
        years_simulated,
        results_enduse_every_year,
        rs_enduses,
        ss_enduses,
        is_enduses,
        fig_name
    ):
    """Plots summarised endues for the three sectors. Annual
    GWh are converted into GW.

    Arguments
    ----------
    data : dict
        Data container
    results_objects :

    enduses_data :

    Note
    ----
        -   Sum across all fueltypes

    # INFO Cannot plot a single year?
    """
    logging.info("plot annual demand for enduses for all submodels")

    submodels = ['residential', 'service', 'industry']
    nr_submodels = len(submodels)

    x_data = years_simulated
    y_data = np.zeros((nr_submodels, len(years_simulated)))

    # Set figure size
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(8, 8))

    ax = fig.add_subplot(1, 1, 1)

    for model_year, data_model_run in enumerate(results_enduse_every_year.values()):

        submodel = 0
        for fueltype_int in range(lookups['fueltypes_nr']):
            for enduse in rs_enduses:

                # Conversion: Convert gwh per years to gw
                yearly_sum_gw = np.sum(data_model_run[enduse][fueltype_int])
                yearly_sum_twh = conversions.gwh_to_twh(yearly_sum_gw)
                y_data[submodel][model_year] += yearly_sum_twh #yearly_sum_gw

        submodel = 1
        for fueltype_int in range(lookups['fueltypes_nr']):
            for enduse in ss_enduses:

                # Conversion: Convert gwh per years to gw
                yearly_sum_gw = np.sum(data_model_run[enduse][fueltype_int])
                yearly_sum_twh = conversions.gwh_to_twh(yearly_sum_gw)
                y_data[submodel][model_year] += yearly_sum_twh #yearly_sum_gw

        submodel = 2
        for fueltype_int in range(lookups['fueltypes_nr']):
            for enduse in is_enduses:

                # Conversion: Convert gwh per years to gw
                yearly_sum_gw = np.sum(data_model_run[enduse][fueltype_int])
                yearly_sum_twh = conversions.gwh_to_twh(yearly_sum_gw)
                y_data[submodel][model_year] += yearly_sum_twh #yearly_sum_gw

    # Convert to stack
    y_stacked = np.row_stack((y_data))

    color_stackplots = ['darkturquoise', 'orange', 'firebrick']

    # ----------
    # Stack plot
    # ----------
    ax.stackplot(
        x_data,
        y_stacked,
        colors=color_stackplots)

    # ------------
    # Plot color legend with colors for every SUBMODEL
    # ------------
    plt.legend(
        submodels,
        ncol=1,
        prop={
            'family': 'arial',
            'size': 5},
        loc='best',
        frameon=False)

    # -------
    # Axis
    # -------
    plt.xticks(years_simulated, years_simulated)
    plt.axis('tight')

    # -------
    # Labels
    # -------
    plt.ylabel("TWh")
    plt.xlabel("year")
    plt.title("UK ED per sector")

    # Tight layout
    plt.margins(x=0)
    fig.tight_layout()

    # Save fig
    plt.savefig(fig_name)
    plt.close()



"""Script to run batch commands
"""
import os
import subprocess
from multiprocessing import Pool, cpu_count

from energy_demand.read_write import read_weather_data

def my_function(simulation_number):  
    print('simulation_number ' + str(simulation_number))

    all_weather_stations = False

    same_weather_yr = True
    defined_weather_yr = 2015

    # --------------------------
    # Get all weather yrs with data (maybe read existing years from data folder)
    # and select weather yr
    # --------------------------
    path_to_weather_data = "C:/Users/cenv0553/ED/data/_raw_data/A-temperature_data/cleaned_weather_stations_data"
    weather_yrs_stations = read_weather_data.get_all_station_per_weather_yr(path_to_weather_data, min_nr_of_stations=30)
    weather_yrs = list(weather_yrs_stations.keys())
    print("all weather yrs:         " + str(weather_yrs))
    print("Total nr of stations:    " + str(len(weather_yrs)))

    if same_weather_yr:
        weather_yr = defined_weather_yr
    else:
        weather_yr = weather_yrs[simulation_number]

    # --------------------------
    # Select weather station
    # --------------------------
    if all_weather_stations:
        weather_station_cnt = [] #All stationssimulation_number
    else:
        weather_station_cnt = simulation_number

    # Run energy demand main.py
    os.system("cd C:/Users/cenv0553/ed")
    os.system("python energy_demand/energy_demand/main.py {} {}".format(weather_yr, weather_station_cnt))

    print("======================")
    print("Finished model run simulation: weather_yr: {} weather_station_cnt: {}".format(weather_yr, weather_station_cnt))
    print("======================")

# ===============================
#simulation_number = range(40)
simulation_number = range(0, 200)

for i in simulation_number:
    my_function(i)



"""All necessary calculations in pre_simulate()
"""
import os
import logging

from energy_demand.read_write import data_loader
from energy_demand.basic import basic_functions
from energy_demand.scripts import init_scripts
from energy_demand.read_write import write_data
from energy_demand.assumptions import general_assumptions
from energy_demand.validation import lad_validation
from energy_demand.scripts import s_disaggregation
from energy_demand.basic import demand_supply_interaction
from energy_demand.read_write import read_data
from energy_demand.technologies import tech_related
from energy_demand.scripts import s_scenario_param
from energy_demand.geography import weather_region
from energy_demand.basic import testing_functions

def load_data_before_simulation(
        data,
        sim_yrs,
        config,
        curr_yr
    ):
    # ---------
    # Configuration
    # -----------
    base_yr = config['CONFIG']['base_yr']
    weather_yr_scenario = config['CONFIG']['weather_yr_scenario']
    path_new_scenario = config['PATHS']['path_new_scenario']

    data['weather_station_count_nr'] = []  # Default value is '[]' to use all stations
    data['data_path'] = os.path.normpath(config['PATHS']['path_local_data'])
    data['processed_path'] = os.path.normpath(config['PATHS']['path_processed_data'])
    data['result_path'] = os.path.normpath(config['PATHS']['path_result_data'])
    data['paths'] = config['CONFIG_DATA']

    # Downloaded (FTP) data
    data['local_paths'] = config['DATA_PATHS']

    # ------------------------------------------------
    # Load Inputs
    # ------------------------------------------------
    data['enduses'], data['sectors'], data['fuels'], lookup_enduses, \
        lookup_sector_enduses = data_loader.load_fuels(config['CONFIG_DATA'])

    # ------------------------------------------------
    # Load Assumptions
    # ------------------------------------------------
    data['assumptions'] = general_assumptions.Assumptions(
        lookup_enduses=lookup_enduses,
        lookup_sector_enduses=lookup_sector_enduses,
        base_yr=base_yr,
        weather_by=config['CONFIG']['user_defined_weather_by'],
        simulation_end_yr=config['CONFIG']['user_defined_simulation_end_yr'],
        curr_yr=curr_yr,
        sim_yrs=sim_yrs,
        paths=config['CONFIG_DATA'],
        enduses=data['enduses'],
        sectors=data['sectors'],
        reg_nrs=len(data['regions']))

    # --------------------------------------------
    # Make selection of weather stations and data
    # --------------------------------------------
    weather_stations_selection = {}
    temp_data_selection = {}
    if data['weather_station_count_nr'] != []:
        for year in sim_yrs:
            weather_stations_selection, station_id = weather_region.get_weather_station_selection(
                data['weather_stations'],
                counter=data['weather_station_count_nr'],
                weather_yr=weather_yr_scenario)
            temp_data_selection[year] = data['temp_data'][year][station_id]

            if year == weather_yr_scenario:
                data['simulation_name'] = str(weather_yr_scenario) + "__" + str(station_id)
    else:
        for year in sim_yrs:
            weather_stations_selection = data['weather_stations']
            temp_data_selection[year] = data['temp_data'][year]

            if year == weather_yr_scenario:
                data['simulation_name'] = str(weather_yr_scenario) + "__" + "all_stations"

    # Replace weather station with selection
    data['weather_stations'] = weather_stations_selection
    data['temp_data'] = temp_data_selection

    # ------------------------------------------
    # Make selection of regions to model
    # ------------------------------------------
    if config['CRITERIA']['reg_selection']:
        
        region_selection = read_data.get_region_selection(
            os.path.join(config['DATA_PATHS']['local_path_datafolder'],
            "region_definitions",
            config['CRITERIA']['reg_selection_csv_name']))
        #region_selection = ['E02003237', 'E02003238']
        setattr(data['assumptions'], 'reg_nrs', len(region_selection))
    else:
        region_selection = data['regions']

    # Create .ini file with simulation parameter
    write_data.write_simulation_inifile(
        path_new_scenario, data, region_selection)

    # -------------------------------------------
    # Weather year specific initialisations
    # -------------------------------------------
    path_folder_weather_yr = os.path.join(
        os.path.join(path_new_scenario, data['simulation_name']))

    data['weather_result_paths'] = basic_functions.get_result_paths(path_folder_weather_yr)

    folders_to_create = [
        path_folder_weather_yr,
        data['weather_result_paths']['data_results'],
        data['weather_result_paths']['data_results_PDF'],
        data['weather_result_paths']['data_results_validation'],
        data['weather_result_paths']['data_results_model_run_results_txt']]
    for folder in folders_to_create:
        basic_functions.create_folder(folder)

    # ------------------------------------------------
    # Load load profiles of technologies
    # ------------------------------------------------
    data['tech_lp'] = data_loader.load_data_profiles(
        config['CONFIG_DATA'],
        config['DATA_PATHS'],
        data['assumptions'].model_yeardays,
        data['assumptions'].model_yeardays_daytype)

    # Obtain population data for disaggregation
    if config['CRITERIA']['msoa_crit']:
        name_population_dataset = config['DATA_PATHS']['path_population_data_for_disaggregation_msoa']
    else:
        name_population_dataset = config['DATA_PATHS']['path_population_data_for_disaggregation_lad']
    data['pop_for_disag'] = data_loader.read_scenario_data(
        name_population_dataset,
        region_name='lad_uk_2016', 
        value_name='population')

    # ------------------------------------------------
    # Load building related data
    # ------------------------------------------------
    if config['CRITERIA']['virtual_building_stock_criteria']:
        data['scenario_data']['floor_area']['rs_floorarea'], data['scenario_data']['floor_area']['ss_floorarea'], data['service_building_count'], rs_regions_without_floorarea, ss_regions_without_floorarea = data_loader.floor_area_virtual_dw(
            data['regions'],
            data['sectors'],
            config['DATA_PATHS'],
            data['scenario_data']['population'][data['assumptions'].base_yr],
            base_yr=data['assumptions'].base_yr)

        # Add all areas with no floor area data
        data['assumptions'].update("rs_regions_without_floorarea", rs_regions_without_floorarea)
        data['assumptions'].update("ss_regions_without_floorarea", ss_regions_without_floorarea)
    else:
        # ------------------------------------------------
        # Load floor area directly from Newcastle
        # ------------------------------------------------
        #rs_floorarea = defaultdict(dict)
        #ss_floorarea = defaultdict(dict)
        pass

    return data

def before_simulation(
        data,
        config,
        sim_yrs,
        pop_density,
        service_switches,
        fuel_switches,
        capacity_switches
    ):
    """
    """
    # ------------------------------------------------
    # Disaggregate national energy demand to regional demands
    # ------------------------------------------------
    fuel_disagg = s_disaggregation.disaggr_demand(
        data,
        config['CRITERIA']['crit_temp_min_max'],
        spatial_calibration=config['CRITERIA']['spatial_calibration'])

    # ------------------------------------------------
    # Calculate spatial diffusion factors
    #
    # Here the real values used for the spatial disaggregation (speec_con_max)
    # need to be defined. If not population density is used,
    # this needs to be replaced by any other values which are loaded from
    # a csv file in the form of: {{region_name: value}}
    # ------------------------------------------------
    f_reg, f_reg_norm, f_reg_norm_abs, crit_all_the_same = init_scripts.create_spatial_diffusion_factors(
        narrative_spatial_explicit_diffusion=data['assumptions'].strategy_vars['spatial_explicit_diffusion'],
        fuel_disagg=fuel_disagg,
        regions=data['regions'],
        real_values=pop_density,
        narrative_speed_con_max=data['assumptions'].strategy_vars['speed_con_max'])

    # ------------------------------------------------
    # Calculate parameter values for every region
    # ------------------------------------------------
    regional_vars = init_scripts.spatial_explicit_modelling_strategy_vars(
        data['assumptions'].strategy_vars,
        data['assumptions'].spatially_modelled_vars,
        data['regions'],
        fuel_disagg,
        f_reg,
        f_reg_norm,
        f_reg_norm_abs)
    data['assumptions'].update('strategy_vars', regional_vars)

    # ------------------------------------------------
    # Calculate parameter values for every simulated year
    # based on narratives. Also calculate annual parameters for
    # technologies diffused by switches.
    # ------------------------------------------------
    regional_vars, non_regional_vars = s_scenario_param.generate_annual_param_vals(
        data['regions'],
        data['assumptions'].strategy_vars,
        sim_yrs)

    # ------------------------------------------------
    # Switches calculations
    # ------------------------------------------------
    # Update assumptions
    crit_switch_happening = testing_functions.switch_testing(
        fuel_switches=fuel_switches,
        service_switches=service_switches,
        capacity_switches=capacity_switches)
    setattr(data['assumptions'],'crit_switch_happening', crit_switch_happening)

    annual_tech_diff_params = init_scripts.switch_calculations(
        sim_yrs,
        data,
        f_reg,
        f_reg_norm,
        f_reg_norm_abs,
        crit_all_the_same,
        service_switches=service_switches,
        fuel_switches=fuel_switches,
        capacity_switches=capacity_switches)

    for region in data['regions']:
        regional_vars[region]['annual_tech_diff_params'] = annual_tech_diff_params[region]

    return regional_vars, non_regional_vars, fuel_disagg, crit_switch_happening

def write_user_defined_results(
        criterias,
        result_paths,
        sim_obj,
        data,
        curr_yr,
        region_selection,
        pop_array_cy
    ):
    """
    Write annual results to files
    """

    logging.info("... Start writing results to file")
    if criterias['write_txt_additional_results']:

        # Write population data to file
        write_data.write_scenaric_population_data(
            curr_yr,
            result_paths['data_results_model_run_pop'],
            pop_array_cy)

        # Write full results (Note: Results in very large data written to file)
        ##write_data.write_full_results(
        ##    data_handle.current_timestep,
        ##    os.path.join(result_path, 'model_run_results_txt'),
        ##    sim_obj.ed_enduse_fueltype_regs_yh,
        ##    "out_enduse_specific")
        write_data.write_supply_results(
            curr_yr,
            "ed_fueltype_regs_yh",
            result_paths['data_results_model_run_pop'],
            sim_obj.ed_fueltype_regs_yh,
            "result_tot_submodels_fueltypes")
        write_data.write_enduse_specific(
            curr_yr,
            result_paths['data_results_model_run_results_txt'],
            sim_obj.tot_fuel_y_enduse_specific_yh,
            "out_enduse_specific")
        write_data.write_lf(
            result_paths['data_results_model_run_results_txt'],
            "result_reg_load_factor_y",
            [curr_yr], sim_obj.reg_load_factor_y, 'reg_load_factor_y')
        write_data.write_lf(
            result_paths['data_results_model_run_results_txt'],
            "result_reg_load_factor_yd",
            [curr_yr], sim_obj.reg_load_factor_yd, 'reg_load_factor_yd')

    # ----------------------------------------------------------------------------------------
    # Write out national demand for every fueltype (used for first sending of demand data)
    # ----------------------------------------------------------------------------------------
    if criterias['write_out_national']:

        # Write out gas
        demand_supply_interaction.write_national_results(
            path_folder=result_paths['data_results'],
            results_unconstrained=sim_obj.results_unconstrained,
            enduse_specific_results=sim_obj.tot_fuel_y_enduse_specific_yh,
            fueltype_str='gas',
            fuelype_nr=tech_related.get_fueltype_int('gas'),
            year=curr_yr,
            submodels_names=data['assumptions'].submodels_names)

        # Write out elec
        demand_supply_interaction.write_national_results(
            path_folder=result_paths['data_results'],
            results_unconstrained=sim_obj.results_unconstrained,
            enduse_specific_results=sim_obj.tot_fuel_y_enduse_specific_yh,
            fueltype_str='electricity',
            fuelype_nr=tech_related.get_fueltype_int('electricity'),
            year=curr_yr,
            submodels_names=data['assumptions'].submodels_names)

    # ------------------------------------------------
    # Temporal Validation
    # ------------------------------------------------
    if (criterias['validation_criteria'] == True) and (
        curr_yr == data['assumptions'].base_yr) and (['cluster_calc'] != True):
        lad_validation.spatio_temporal_val(
            sim_obj.ed_fueltype_national_yh,
            sim_obj.ed_fueltype_regs_yh,
            result_paths,
            data['paths'],
            region_selection,
            data['assumptions'].seasons,
            data['assumptions'].model_yeardays_daytype,
            plot_crit=False)

def load_gva_sector(
        data_handle,
        regions,
        sectors_to_load,
        msoa_crit,
        simulate=False
    ):
    """Load sector specific GVA

    Arguments
    ---------
    data_handle : object
        Data handler
    pop_array : array
        Population
    regions : list
        Regions
    sectors_to_load : list
        Sectors which are loaded
    msoa_crit : bool
        Criteria wheter modelled on MSOA level or LAD level
    simulate : bool
        Criteria wheter run in simulate() or not
    """
    sector_data = {}
    if msoa_crit:
        logging.info("Don't load sector GVA {}")
    else:
        for gva_sector_nr in sectors_to_load:
            try:
                logging.info("... Loading GVA data for sector_Nr {}".format(gva_sector_nr))
                if simulate:
                    gva_sector_data = data_handle.get_data(
                        'gva_per_head_sector__{}'.format(gva_sector_nr))
                else:
                    gva_sector_data = data_handle.get_base_timestep_data(
                        'gva_per_head_sector__{}'.format(gva_sector_nr))

                sector_data[gva_sector_nr] = basic_functions.assign_array_to_dict(
                    gva_sector_data, regions)
            except KeyError:
                # In case no data could be loaded, generate constant dummy data
                raise Exception("Could not load data %s", 'gva_per_head_sector__{}'.format(gva_sector_nr))

    return sector_data

def plots(
        data,
        curr_yr,
        fuel_disagg, 
        config
    ):
    """
    """
    # Spatial validation
    if (config['CRITERIA']['validation_criteria'] == True) and (
        curr_yr == data['assumptions'].base_yr) and (
            config['CRITERIA']['cluster_calc'] != True):
        lad_validation.spatial_validation_lad_level(
            fuel_disagg,
            data['result_paths'],
            data['paths'],
            data['regions'],
            data['reg_coord'],
            plot_crit=False)

    # Plot map with weather station
    if config['CRITERIA']['cluster_calc'] != True:
        data_loader.create_weather_station_map(
            data['weather_stations'],
            os.path.join(data['result_path'], 'weatherst_distr_weathyr_{}.pdf'.format(
                config['CONFIG']['weather_yr_scenario'])),
            path_shapefile=config['DATA_PATHS']['lad_shapefile'])



"""Plot peak and total demand for 
"""
import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from energy_demand.basic import conversions
from energy_demand.technologies import tech_related
from energy_demand.plotting import basic_plot_functions
from energy_demand.read_write import write_data

def run(
        data_input,
        fueltype_str,
        fig_name
    ):
    """Plot peak demand and total demand over time in same plot
    """
    statistics_to_print = []

    # Select period and fueltype
    fueltype_int = tech_related.get_fueltype_int(fueltype_str)

    # -----------------------------------------------------------
    # Modelled years
    # -----------------------------------------------------------

    # List of selected data for every weather year (which is then converted to array)
    weather_yrs_total_demand = []
    weather_yrs_peak_demand = []

    nr_weather_yrs = list(data_input.keys())
    statistics_to_print.append("_____________________________")
    statistics_to_print.append("Weather years")
    statistics_to_print.append(str(data_input.keys()))

    # Iterate weather years
    for weather_yr, data_weather_yr in data_input.items():

        total_demands = []
        peak_demands = []
        sim_yrs = []
        for sim_yr in data_weather_yr.keys():
            sim_yrs.append(sim_yr)
            data_input_fueltype = data_weather_yr[sim_yr][fueltype_int]     # Select fueltype

            # sum total annual demand and convert gwh to twh
            sum_gwh_y = np.sum(data_input_fueltype)
            sum_thw_y = conversions.gwh_to_twh(sum_gwh_y)

            # Get peak
            peak_h = np.max(data_input_fueltype.reshape(8760))

            total_demands.append(sum_thw_y)
            peak_demands.append(peak_h)

        weather_yrs_total_demand.append(total_demands)
        weather_yrs_peak_demand.append(peak_demands)

    columns = sim_yrs

    # Convert to array
    weather_yrs_total_demand = np.array(weather_yrs_total_demand)
    weather_yrs_peak_demand = np.array(weather_yrs_peak_demand)

    # Calculate std per simulation year
    std_total_demand = list(np.std(weather_yrs_total_demand, axis=0)) # across columns calculate std
    std_peak_demand = list(np.std(weather_yrs_peak_demand, axis=0)) # across columns calculate std

    # Create dataframe
    if len(nr_weather_yrs) > 2:

        # Create dataframes
        df_total_demand = pd.DataFrame(weather_yrs_total_demand, columns=columns)
        df_peak = pd.DataFrame(weather_yrs_peak_demand, columns=columns)

        # Calculate quantiles
        quantile_95 = 0.95
        quantile_05 = 0.05

        # Calculate quantiles
        df_total_demand_q_95 = df_total_demand.quantile(quantile_95)
        df_total_demand_q_05 = df_total_demand.quantile(quantile_05)
        df_peak_q_95 = df_peak.quantile(quantile_95)
        df_peak_q_05 = df_peak.quantile(quantile_05)

        # convert to list
        df_total_demand_q_95 = df_total_demand_q_95.tolist()
        df_total_demand_q_05 = df_total_demand_q_05.tolist()
        df_peak_q_95 = df_peak_q_95.tolist()
        df_peak_q_05 = df_peak_q_05.tolist()
        #df_peak = df_peak.T #All indivdiual values
    else:
        #df_total_demand = weather_yrs_total_demand
        #df_peak = weather_yrs_peak_demand
        pass

    # -------------------
    # Base year data (2015)
    # -------------------
    # total demand
    tot_demand_twh_2015 = []
    for sim_yr, data_sim_yr in data_input[2015].items():
        gwh_2015_y = np.sum(data_sim_yr[fueltype_int])
        twh_2015_y = conversions.gwh_to_twh(gwh_2015_y)
        tot_demand_twh_2015.append(twh_2015_y)

    # peak
    df_peak_2015 = []
    for sim_yr, data_sim_yr in data_input[2015].items():
        peak_gwh_2015_y = np.max(data_sim_yr[fueltype_int])
        df_peak_2015.append(peak_gwh_2015_y)

    # ---------------
    # Smoothing lines
    # ---------------
    if len(nr_weather_yrs) > 2:
        try:
            period_h_smoothed, tot_demand_twh_2015_smoothed = basic_plot_functions.smooth_data(columns, tot_demand_twh_2015, num=40000)
            period_h_smoothed, df_total_demand_q_95_smoothed = basic_plot_functions.smooth_data(list(columns), df_total_demand_q_95, num=40000)
            period_h_smoothed, df_total_demand_q_05_smoothed = basic_plot_functions.smooth_data(columns, df_total_demand_q_05, num=40000)
            period_h_smoothed, df_peak_q_95_smoothed = basic_plot_functions.smooth_data(list(columns), df_peak_q_95, num=40000)
            period_h_smoothed, df_peak_q_05_smoothed = basic_plot_functions.smooth_data(columns, df_peak_q_05, num=40000)
            period_h_smoothed, df_peak_2015_smoothed = basic_plot_functions.smooth_data(columns, df_peak_2015, num=40000)
        except:
            period_h_smoothed = columns
            df_total_demand_q_95_smoothed = df_total_demand_q_95
            df_total_demand_q_05_smoothed = df_total_demand_q_05
            df_peak_q_95_smoothed = df_peak_q_95
            df_peak_q_05_smoothed = df_peak_q_05
            tot_demand_twh_2015_smoothed = tot_demand_twh_2015
            df_peak_2015_smoothed = df_peak_2015
    else:
        try:
            period_h_smoothed, tot_demand_twh_2015_smoothed = basic_plot_functions.smooth_data(columns, tot_demand_twh_2015, num=40000)
            period_h_smoothed, df_peak_2015_smoothed = basic_plot_functions.smooth_data(columns, df_peak_2015, num=40000)
        except:
            period_h_smoothed = columns
            tot_demand_twh_2015_smoothed = tot_demand_twh_2015
            df_peak_2015_smoothed = df_peak_2015
     
    # --------------
    # Two axis figure
    # --------------
    fig, ax1 = plt.subplots(
        figsize=basic_plot_functions.cm2inch(15, 10))

    ax2 = ax1.twinx()

    # Axis label
    ax1.set_xlabel('Years')
    ax2.set_ylabel('Peak hour {} demand (GW)'.format(fueltype_str), color='black')
    ax1.set_ylabel('Total {} demand (TWh)'.format(fueltype_str), color='black')

    # Make the y-axis label, ticks and tick labels match the line color.¨
    color_axis1 = 'lightgrey'
    color_axis2 = 'blue'

    ax1.tick_params('y', colors='black')
    ax2.tick_params('y', colors='black')

    if len(nr_weather_yrs) > 2:

        # -----------------
        # Uncertainty range total demand
        # -----------------
        '''ax1.plot(
            period_h_smoothed,
            df_total_demand_q_05_smoothed,
            color='tomato', linestyle='--', linewidth=0.5, label="0.05_total_demand")'''

        '''ax1.plot(
            period_h_smoothed,
            df_total_demand_q_95_smoothed,
            color=color_axis1, linestyle='--', linewidth=0.5, label="0.95_total_demand")

        ax1.fill_between(
            period_h_smoothed, #x
            df_total_demand_q_95_smoothed,  #y1
            df_total_demand_q_05_smoothed,  #y2
            alpha=.25,
            facecolor=color_axis1,
            label="uncertainty band demand")'''

        # -----------------
        # Uncertainty range peaks
        # -----------------
        ##ax2.plot(period_h_smoothed, df_peak_q_05_smoothed, color=color_axis2, linestyle='--', linewidth=0.5, label="0.05_peak")
        ##ax2.plot(period_h_smoothed, df_peak_q_95_smoothed, color=color_axis2, linestyle='--', linewidth=0.5, label="0.95_peak")
        ax2.plot(
            period_h_smoothed,
            df_peak_2015_smoothed,
            color=color_axis2, linestyle="--", linewidth=0.4)

        # Error bar of bar charts
        ax2.errorbar(columns, df_peak_2015, linewidth=0.5, color='black', yerr=std_peak_demand, linestyle="None")

        # Error bar bar plots
        ax1.errorbar(
            columns, tot_demand_twh_2015, linewidth=0.5, color='black', yerr=std_total_demand, linestyle="None")

        '''ax2.fill_between(
            period_h_smoothed, #x
            df_peak_q_95_smoothed,  #y1
            df_peak_q_05_smoothed,  #y2
            alpha=.25,
            facecolor="blue",
            label="uncertainty band peak")'''

    # Total demand bar plots
    ##ax1.plot(period_h_smoothed, tot_demand_twh_2015_smoothed, color='tomato', linestyle='-', linewidth=2, label="tot_demand_weather_yr_2015")
    ax1.bar(
        columns,
        tot_demand_twh_2015,
        width=2,
        alpha=1,
        align='center',
        color=color_axis1,
        label="total {} demand".format(fueltype_str))

    statistics_to_print.append("_____________________________")
    statistics_to_print.append("total demand per model year")
    statistics_to_print.append(str(tot_demand_twh_2015))

    # Line of peak demand
    #ax2.plot(columns, df_peak, color=color_axis2, linestyle='--', linewidth=0.5, label="peak_0.95")
    ax2.plot(
        period_h_smoothed,
        df_peak_2015_smoothed,
        color=color_axis2, linestyle='-', linewidth=2, label="{} peak demand (base weather yr)".format(fueltype_str))

    statistics_to_print.append("_____________________________")
    statistics_to_print.append("peak demand per model year")
    statistics_to_print.append(str(df_peak_2015))

    # Scatter plots of peak demand
    ax2.scatter(
        columns,
        df_peak_2015,
        marker='o', s=20, color=color_axis2, alpha=1)

    ax1.legend(
        prop={
            'family':'arial',
            'size': 10},
        loc='upper center',
        bbox_to_anchor=(0.9, -0.1),
        frameon=False,
        shadow=True)

    ax2.legend(
        prop={
            'family':'arial',
            'size': 10},
        loc='upper center',
        bbox_to_anchor=(0.1, -0.1),
        frameon=False,
        shadow=True)

    # More space at bottom
    #fig.subplots_adjust(bottom=0.4)
    fig.tight_layout()

    plt.savefig(fig_name)
    plt.close()

    # Write info to txt
    write_data.write_list_to_txt(
        os.path.join(fig_name.replace(".pdf", ".txt")),
        statistics_to_print)
    print("--")



"""Weather Station location
"""
from haversine import haversine

def calc_distance_two_points(lat_from, long_from, lat_to, long_to):
    """Calculate distance between two points

    https://pypi.org/project/haversine/#description

    Arguments
    ----------
    long_from : float
        Longitute coordinate from point
    lat_from : float
        Latitute coordinate from point
    long_to : float
        Longitute coordinate to point
    lat_to : float
        Latitue coordinate to point

    Return
    ------
    distance : float
        Distance
    """
    distance_in_km = haversine(
        (lat_from, long_from),
        (lat_to, long_to),
        unit='km')

    return distance_in_km

def get_closest_station(
        latitude_reg,
        longitude_reg,
        weather_stations
    ):
    """Search ID of closest weater station

    Arguments
    ----------
    longitude_reg : float
        Longitute coordinate of Region Object
    latitude_reg : float
        Latitute coordinate of Region Object
    weather_stations : dict
        Weater station data

    Return
    ------
    closest_id : int
        ID of closest weather station
    """
    closest_dist = 99999999999

    for station_id in weather_stations:

        try:
            lat_to = weather_stations[station_id].latitude
            long_to = weather_stations[station_id].longitude
        except AttributeError:
            lat_to = weather_stations[station_id]["latitude"]
            long_to = weather_stations[station_id]["longitude"]

        dist_to_station = calc_distance_two_points(
            lat_from=latitude_reg,
            long_from=longitude_reg,
            lat_to=lat_to,
            long_to=long_to)

        if dist_to_station < closest_dist:
            closest_dist = dist_to_station
            closest_id = station_id

    return closest_id



"""
"""
import numpy as np
from energy_demand.technologies import technological_stock

def test_Technology():
    """
    """
    tech = technological_stock.Technology(
        name="boilerA",
        tech_type="heating_tech",
        fueltype_str="electricity",
        eff_achieved=1.0,
        diff_method='linear',
        eff_by=0.5,
        eff_ey=1.0,
        year_eff_ey=2020,
        base_yr=2015,
        curr_yr=2020,
        temp_by=np.zeros((365, 24)) + 10,
        temp_cy=np.zeros((365, 24)) + 10,
        t_base_heating_by=15.5,
        t_base_heating_cy=15.5)

    assert tech.name == "boilerA"
    assert tech.eff_cy == 1.0

    tech2 = technological_stock.Technology(
        name="boilerA",
        tech_type="heat_pump",
        fueltype_str="electricity",
        eff_achieved=1.0,
        diff_method='linear',
        eff_by=0.5,
        eff_ey=1.0,
        year_eff_ey=2020,
        base_yr=2015,
        curr_yr=2020,
        temp_by=np.zeros((365, 24)) + 10,
        temp_cy=np.zeros((365, 24)) + 10,
        t_base_heating_by=20,
        t_base_heating_cy=20)

    assert tech2.eff_cy == 1.0

    tech3 = technological_stock.Technology(
        name="dummy_tech",
        tech_type="heating_tech")

    assert tech3.name == "dummy_tech"

def test_TechStock():

    from energy_demand.read_write import read_data

    base_yr = 2015
    curr_yr = 2020

    all_technologies = {'boilerA': read_data.TechnologyData()}
    all_technologies['boilerA'].fueltype_str = 'electricity'
    all_technologies['boilerA'].eff_achieved = 1.0
    all_technologies['boilerA'].diff_method = 'linear'
    all_technologies['boilerA'].eff_by = 1.0
    all_technologies['boilerA'].eff_ey = 1.0
    all_technologies['boilerA'].year_eff_ey = 2020

    stock_obj = technological_stock.TechStock(
        name="name",
        technologies=all_technologies,
        base_yr=base_yr,
        curr_yr=curr_yr,
        fueltypes={'electricity': 2},
        temp_by=np.ones((365, 24)) + 10,
        temp_cy=np.ones((365, 24)) + 10,
        t_base_heating_by=15.5,
        potential_enduses=['heating'],
        t_base_heating_cy=15.5,
        enduse_technologies={'heating': {'sectorA': ['boilerA']}})

    assert stock_obj.name == "name"
    assert stock_obj.get_tech_attr('heating', 'sectorA', 'boilerA', 'eff_by') == 1.0



"""TEsting functions
"""
import numpy as np

from energy_demand.basic import basic_functions
from energy_demand.basic import lookup_tables

def test_if_minus_value_in_array(arraytotest):#, tolerance_min_max=0.00000000001):
    """Test if array has negative value according to a tolerance
    criteria

    Arguments
    ---------
    arraytotest : array
        Input to test for negative values
    tolerance_min_max : float
        Minimum and maximum tolerance criteria

    Returns
    -------
    bool : Whether empty number in array or not
    """

    only_neg_elements = arraytotest[arraytotest < 0]

    if len(only_neg_elements) > 0:
        #for element in only_neg_elements:
            #if (element > tolerance_min_max) or (element < tolerance_min_max):
            #if element < 0:
        print("---")
        print("Sum of all negative:     " + str(np.sum(only_neg_elements)))
        print("Average negative value:  " + str(np.sum(only_neg_elements) / len(only_neg_elements)))
        print("Smalles value:           " + str(np.min(only_neg_elements)))
        return True
    else:
        return False

def switch_testing(fuel_switches, service_switches, capacity_switches):
    """Test if swithes defined for same enduse

    Arguments
    ---------
    fuel_switches : list
        Switches
    service_switches : list
        Switches
    capacity_switches : list
        Switches
    """
    all_switches_incl_sectors = {}

    enduses_service_switch = set([])

    for switch in service_switches:
        enduses_service_switch.add(switch.enduse)

        # Collect all enduses and affected sectors
        if switch.enduse not in all_switches_incl_sectors:
            all_switches_incl_sectors[switch.enduse] = set([])

            if not switch.sector:
                all_switches_incl_sectors[switch.enduse] = None
            else:
                all_switches_incl_sectors[switch.enduse].add(switch.sector)
        else:
            if not switch.sector:
                pass
            else:
                all_switches_incl_sectors[switch.enduse].add(switch.sector)

    enduses_capacity_switch = set([])

    for switch in capacity_switches:
        enduses_capacity_switch.add(switch.enduse)

        # Collect all enduses and affected sectors
        if switch.enduse not in all_switches_incl_sectors:
            all_switches_incl_sectors[switch.enduse] = set([])

            if not switch.sector:
                all_switches_incl_sectors[switch.enduse] = None
            else:
                all_switches_incl_sectors[switch.enduse].add(switch.sector)
        else:
            if not switch.sector:
                pass
            else:
                all_switches_incl_sectors[switch.enduse].add(switch.sector)

    enduses_service_switch = list(enduses_service_switch)
    enduses_capacity_switch = list(enduses_capacity_switch)

    for enduse in all_switches_incl_sectors:
        if all_switches_incl_sectors[enduse] != None:
            all_switches_incl_sectors[enduse] = list(all_switches_incl_sectors[enduse])

    return all_switches_incl_sectors

def testing_fuel_tech_shares(fuel_tech_fueltype_p):
    """Test if assigned fuel share add up to 1 within each fueltype

    Paramteres
    ----------
    fuel_tech_fueltype_p : dict
        Fueltype fraction of technologies
    """
    for enduse in fuel_tech_fueltype_p:

        sector_crit = basic_functions.test_if_sector(
            fuel_tech_fueltype_p[enduse])

        if sector_crit:
            for sector in fuel_tech_fueltype_p[enduse]:
                for fueltype in fuel_tech_fueltype_p[enduse][sector]:
                    if fuel_tech_fueltype_p[enduse][sector][fueltype] != {}:
                        if round(sum(fuel_tech_fueltype_p[enduse][sector][fueltype].values()), 3) != 1.0:
                            raise Exception(
                                "The fuel shares assumptions are wrong for enduse {} and fueltype {} SUM: {}".format(
                                    enduse, fueltype, sum(fuel_tech_fueltype_p[enduse][sector][fueltype].values())))
        else:
            for fueltype in fuel_tech_fueltype_p[enduse]:
                if fuel_tech_fueltype_p[enduse][fueltype] != {}:
                    if round(sum(fuel_tech_fueltype_p[enduse][fueltype].values()), 3) != 1.0:
                        raise Exception(
                            "The fuel shares assumptions are wrong for enduse {} and fueltype {} SUM: {}".format(
                                enduse, fueltype, sum(fuel_tech_fueltype_p[enduse][fueltype].values())))

def testing_tech_defined(technologies, all_tech_enduse):
    """Test if all technologies are defined for assigned fuels

    Arguments
    ----------
    technologies : dict
        Technologies
    all_tech_enduse : dict
        All technologies per enduse with assigned fuel shares
    """
    for enduse in all_tech_enduse:
        for sector in all_tech_enduse[enduse]:
            for tech in all_tech_enduse[enduse][sector]:
                if tech not in technologies:
                    raise Exception(
                        "Error: '{}' is not defined in technology_definition.csv".format(
                            tech))

def test_function_fuel_sum(
        data,
        fuel_disagg,
        mode_constrained,
        space_heating_enduses
    ):
    """ Sum raw disaggregated fuel data
    """
    lookups = lookup_tables.basic_lookups()
    fuel_in = 0
    fuel_in_solid_fuel = 0
    fuel_in_gas = 0
    fuel_in_elec = 0
    fuel_in_oil = 0
    fuel_in_heat = 0
    fuel_in_hydrogen = 0
    fuel_in_biomass = 0
    tot_heating = 0
    dummy_sector = None
    for region in fuel_disagg['residential']:
        for enduse in fuel_disagg['residential'][region]:
            fuel_in += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector])
            fuel_in_heat += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector][lookups['fueltypes']['heat']])

            if mode_constrained == False and enduse in space_heating_enduses: #Exclude inputs for heating
                tot_heating += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector])
            else:
                fuel_in_elec += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector][lookups['fueltypes']['electricity']])
                fuel_in_gas += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector][lookups['fueltypes']['gas']])
                fuel_in_hydrogen += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector][lookups['fueltypes']['hydrogen']])
                fuel_in_oil += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector][lookups['fueltypes']['oil']])
                fuel_in_solid_fuel += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector][lookups['fueltypes']['solid_fuel']])
                fuel_in_biomass += np.sum(fuel_disagg['residential'][region][enduse][dummy_sector][lookups['fueltypes']['biomass']])

    for region in fuel_disagg['service']:
        for enduse in fuel_disagg['service'][region]:
            for sector in fuel_disagg['service'][region][enduse]:
                fuel_in += np.sum(fuel_disagg['service'][region][enduse][sector])
                fuel_in_heat += np.sum(fuel_disagg['service'][region][enduse][sector][lookups['fueltypes']['heat']])

                if mode_constrained == False and enduse in space_heating_enduses:
                    tot_heating += np.sum(fuel_disagg['service'][region][enduse][sector])
                else:
                    fuel_in_elec += np.sum(fuel_disagg['service'][region][enduse][sector][lookups['fueltypes']['electricity']])
                    fuel_in_gas += np.sum(fuel_disagg['service'][region][enduse][sector][lookups['fueltypes']['gas']])
                    fuel_in_hydrogen += np.sum(fuel_disagg['service'][region][enduse][sector][lookups['fueltypes']['hydrogen']])
                    fuel_in_oil += np.sum(fuel_disagg['service'][region][enduse][sector][lookups['fueltypes']['oil']])
                    fuel_in_solid_fuel += np.sum(fuel_disagg['service'][region][enduse][sector][lookups['fueltypes']['solid_fuel']])
                    fuel_in_biomass += np.sum(fuel_disagg['service'][region][enduse][sector][lookups['fueltypes']['biomass']])
                
    for region in fuel_disagg['industry']:
        for enduse in fuel_disagg['industry'][region]:
            for sector in fuel_disagg['industry'][region][enduse]:
                fuel_in += np.sum(fuel_disagg['industry'][region][enduse][sector])
                fuel_in_heat += np.sum(fuel_disagg['industry'][region][enduse][sector][lookups['fueltypes']['heat']])

                if mode_constrained == False and enduse in space_heating_enduses:
                    tot_heating += np.sum(fuel_disagg['industry'][region][enduse][sector])
                else:
                    fuel_in_elec += np.sum(fuel_disagg['industry'][region][enduse][sector][lookups['fueltypes']['electricity']])
                    fuel_in_gas += np.sum(fuel_disagg['industry'][region][enduse][sector][lookups['fueltypes']['gas']])
                    fuel_in_hydrogen += np.sum(fuel_disagg['industry'][region][enduse][sector][lookups['fueltypes']['hydrogen']])
                    fuel_in_oil += np.sum(fuel_disagg['industry'][region][enduse][sector][lookups['fueltypes']['oil']])
                    fuel_in_solid_fuel += np.sum(fuel_disagg['industry'][region][enduse][sector][lookups['fueltypes']['solid_fuel']])
                    fuel_in_biomass += np.sum(fuel_disagg['industry'][region][enduse][sector][lookups['fueltypes']['biomass']])

    out_dict = {
        "fuel_in": fuel_in,
        "fuel_in_biomass": fuel_in_biomass,
        "fuel_in_elec": fuel_in_elec,
        "fuel_in_gas": fuel_in_gas,
        "fuel_in_heat": fuel_in_heat,
        "fuel_in_hydrogen": fuel_in_hydrogen,
        "fuel_in_solid_fuel": fuel_in_solid_fuel,
        "fuel_in_oil": fuel_in_oil,
        "tot_heating": tot_heating}

    return out_dict

def control_disaggregation(fuel_disagg, national_fuel, enduses, sectors):
    """Check if disaggregation is correct

    Arguments
    ---------
    fuel_disagg : dict
        Disaggregated fuel to regions
    national_fuel : dict
        National fuel
    enduses : list
        Enduses
    sectors : list, bool=False
        Sectors
    """
    control_sum_disagg, control_sum_national = 0, 0

    for reg in fuel_disagg:
        for enduse in fuel_disagg[reg]:
            for sector in fuel_disagg[reg][enduse]:
                control_sum_disagg += np.sum(fuel_disagg[reg][enduse][sector])

    for sector in sectors:
        for enduse in enduses:
            control_sum_national += np.sum(national_fuel[enduse][sector])

    #The loaded floor area must correspond to provided fuel sectors numers
    np.testing.assert_almost_equal(
        control_sum_disagg,
        control_sum_national,
        decimal=2, err_msg="disagregation error ss {} {}".format(
            control_sum_disagg, control_sum_national))
